from bs4 import BeautifulSoup   
from usdm_excel.document.soa import SoA
from usdm_model.activity import Activity
from usdm_model.study_epoch import StudyEpoch
from usdm_model.study_cell import StudyCell
from usdm_model.study_arm import StudyArm
from usdm_model.encounter import Encounter
from usdm_model.study_design import StudyDesign
from usdm_model.schedule_timeline import ScheduleTimeline
from usdm_model.scheduled_instance import ScheduledActivityInstance
from usdm_model.schedule_timeline_exit import ScheduleTimelineExit
from usdm_model.timing import Timing
from usdm_model.condition import Condition

def translate_reference(text, globals):
  soup = BeautifulSoup(str(text), 'html.parser')
  for ref in soup(['usdm:ref']):
    attributes = ref.attrs
    instance = globals.cross_references.get_by_id(attributes['klass'], attributes['id'])
    value = str(getattr(instance, attributes['attribute']))
    ref.replace_with(value)
  return str(soup)

def double_link(items, prev, next):
  for idx, item in enumerate(items):
    if idx == 0:
      setattr(item, prev, None)
    else:
      the_id = getattr(items[idx-1], 'id')
      setattr(item, prev, the_id)
    if idx == len(items)-1:  
      setattr(item, next, None)
    else:
      the_id = getattr(items[idx+1], 'id')
      setattr(item, next, the_id)

def add_cross_ref(collection, globals):
  for item in collection:
    globals.cross_references.add(item.id, item)

def create_conditions(factory, globals):
  item_list = [
    {'name': 'COND1', 'label': '', 'description': '', 'text': 'Only perform at baseline', 'appliesToIds': [], 'contextIds': []},
    {'name': 'COND2', 'label': '', 'description': '', 'text': 'Only perform on males', 'appliesToIds': [], 'contextIds': []},
  ]
  results = factory.set(Condition, item_list)
  add_cross_ref(results, globals)
  return results

def create_timings(factory, globals):
  FIXED = factory.cdisc_code('C201358', 'Fixed Reference')
  BEFORE = factory.cdisc_code('C201356', 'After')
  AFTER = factory.cdisc_code('C201357', 'Before')

  E2E = factory.cdisc_code('C201352', 'End to End')
  E2S = factory.cdisc_code('C201353', 'End to Start')
  S2E = factory.cdisc_code('C201354', 'Start to End')
  S2S = factory.cdisc_code('C201355', 'Start to Start')

  item_list = [
    {'name': 'T1', 'label': '-2 Days', 'description': '', 'type': BEFORE, 'value': '', 'valueLabel': '', 'relativeToFrom': S2S, 'relativeFromScheduledInstanceId': '', 'relativeToScheduledInstanceId': '', 'windowLower': '', 'windowUpper': '', 'windowLabel': ''},
    {'name': 'T2', 'label': 'Dose',    'description': '', 'type': FIXED,  'value': '', 'valueLabel': '', 'relativeToFrom': S2S, 'relativeFromScheduledInstanceId': '', 'relativeToScheduledInstanceId': '', 'windowLower': '', 'windowUpper': '', 'windowLabel': ''},
    {'name': 'T3', 'label': '7 Days',  'description': '', 'type': AFTER,  'value': '', 'valueLabel': '', 'relativeToFrom': S2S, 'relativeFromScheduledInstanceId': '', 'relativeToScheduledInstanceId': '', 'windowLower': '', 'windowUpper': '', 'windowLabel': '1..1 Days'}
  ]
  results = factory.set(Timing, item_list)
  add_cross_ref(results, globals)
  return results

def create_activities(factory, globals):
  item_list = [
    {'name': 'A1', 'label': 'Activity 1', 'description': '', 'definedProcedures': [], 'biomedicalConceptIds': [], 'bcCategoryIds': [], 'bcSurrogateIds': [], 'timelineId': None},
    {'name': 'A2', 'label': 'Activity 2', 'description': '', 'definedProcedures': [], 'biomedicalConceptIds': [], 'bcCategoryIds': [], 'bcSurrogateIds': [], 'timelineId': None},
    {'name': 'A3', 'label': 'Activity 3', 'description': '', 'definedProcedures': [], 'biomedicalConceptIds': [], 'bcCategoryIds': [], 'bcSurrogateIds': [], 'timelineId': None},
    {'name': 'A4', 'label': 'Activity 4', 'description': '', 'definedProcedures': [], 'biomedicalConceptIds': [], 'bcCategoryIds': [], 'bcSurrogateIds': [], 'timelineId': None},
    {'name': 'A5', 'label': 'Activity 5', 'description': '', 'definedProcedures': [], 'biomedicalConceptIds': [], 'bcCategoryIds': [], 'bcSurrogateIds': [], 'timelineId': None}
  ]
  results = factory.set(Activity, item_list)
  double_link(results, 'previousId', 'nextId')
  add_cross_ref(results, globals)
  return results

def create_epochs(factory, globals):
  item_list = [
    {'name': 'EP1', 'label': 'Epoch A', 'description': '', 'type': factory.cdisc_dummy()},
    {'name': 'EP2', 'label': 'Epoch B', 'description': '', 'type': factory.cdisc_dummy()},
    {'name': 'EP3', 'label': 'Epoch C', 'description': '', 'type': factory.cdisc_dummy()},
  ]
  results = factory.set(StudyEpoch, item_list)
  double_link(results, 'previousId', 'nextId')
  add_cross_ref(results, globals)
  return results

def create_encounters(factory, globals):
  item_list = [
    {'name': 'E1', 'label': 'Screening', 'description': '', 'type': factory.cdisc_dummy(), 'environmentalSetting': [], 'contactModes': [], 'transitionStartRule': None, 'transitionEndRule': None, 'scheduledAtId': None},
    {'name': 'E2', 'label': 'Dose', 'description': '', 'type': factory.cdisc_dummy(), 'environmentalSetting': [], 'contactModes': [], 'transitionStartRule': None, 'transitionEndRule': None, 'scheduledAtId': None},
    {'name': 'E3', 'label': 'Check Up', 'description': '', 'type': factory.cdisc_dummy(), 'environmentalSetting': [], 'contactModes': [], 'transitionStartRule': None, 'transitionEndRule': None, 'scheduledAtId': None}
  ]
  results = factory.set(Encounter, item_list)
  double_link(results, 'previousId', 'nextId')
  add_cross_ref(results, globals)
  return results

def create_activity_instances(factory, globals):
  item_list = [
    {'name': 'SAI_1', 'description': '', 'label': '', 'timelineExitId': None, 'encounterId': None, 'scheduledInstanceTimelineId': None, 'defaultConditionId': None, 'epochId': None, 'activityIds': []},
    {'name': 'SAI_2', 'description': '', 'label': '', 'timelineExitId': None, 'encounterId': None, 'scheduledInstanceTimelineId': None, 'defaultConditionId': None, 'epochId': None, 'activityIds': []},
    {'name': 'SAI_3', 'description': '', 'label': '', 'timelineExitId': None, 'encounterId': None, 'scheduledInstanceTimelineId': None, 'defaultConditionId': None, 'epochId': None, 'activityIds': []}
  ]
  results = factory.set(ScheduledActivityInstance, item_list)
  results[0].defaultConditionId = results[1].id
  results[1].defaultConditionId = results[2].id
  add_cross_ref(results, globals)
  return results

def scenario_1(factory, globals):
  dummy_cell = factory.item(StudyCell, {'armId': "X", 'epochId': "Y"})
  dummy_arm = factory.item(StudyArm, {'name': "Arm1", 'type': factory.cdisc_dummy(), 'dataOriginDescription': 'xxx', 'dataOriginType': factory.cdisc_dummy()})
  activities = create_activities(factory, globals)
  epochs = create_epochs(factory, globals)
  encounters = create_encounters(factory, globals)
  activity_instances = create_activity_instances(factory, globals)
  timings = create_timings(factory, globals)
  conditions = create_conditions(factory, globals)

  activity_instances[0].activityIds = [activities[0].id, activities[1].id]
  activity_instances[0].encounterId = encounters[0].id
  activity_instances[0].epochId = epochs[0].id
  activity_instances[1].activityIds = [activities[1].id, activities[2].id]
  activity_instances[1].encounterId = encounters[1].id
  activity_instances[1].epochId = epochs[1].id
  activity_instances[2].activityIds = [activities[2].id, activities[3].id, activities[4].id]
  activity_instances[2].encounterId = encounters[2].id
  activity_instances[2].epochId = epochs[2].id
  timings[0].relativeFromScheduledInstanceId = activity_instances[0].id
  timings[0].relativeToScheduledInstanceId = activity_instances[1].id
  timings[1].relativeFromScheduledInstanceId = activity_instances[1].id
  timings[1].relativeToScheduledInstanceId = activity_instances[1].id
  timings[2].relativeFromScheduledInstanceId = activity_instances[2].id
  timings[2].relativeToScheduledInstanceId = activity_instances[1].id
  conditions[0].appliesToIds = [activities[0].id]
  conditions[0].contextIds = [activity_instances[0].id]
  conditions[1].appliesToIds = [activities[3].id]
  
  exit = factory.item(ScheduleTimelineExit, {})
  activity_instances[-1].timelineExitId = exit.id
  timeline = factory.item(ScheduleTimeline, {'name': 'Study Design', 'label': '', 'description': '', 'mainTimeline': True, 'entryCondition': "Condition", 'entryId': activity_instances[0].id, 'exits': [exit], 'instances': activity_instances, 'timings': timings})
  study_design = factory.item(StudyDesign, {'name': 'Study Design', 'label': '', 'description': '', 
    'rationale': 'XXX', 'interventionModel': factory.cdisc_dummy(),
    'arms': [dummy_arm], 'studyCells': [dummy_cell], 'epochs': epochs,
    'activities': activities, 'scheduledTimelines': [timeline],
    'conditions': conditions})
  return study_design, timeline

def test_create(mocker, globals, factory):
  bs = factory.base_sheet(mocker)
  study_design, timeline = scenario_1(factory, globals)
  soa = SoA(bs, study_design, timeline)
  result = soa.generate()
  #print(f"RESULT: {result}")
  labels = []
  for row in range(len(result)):
    labels.append([])
    for col in range(len(result[row])):
      if 'set' in result[row][col].keys():
        label = 'X' if result[row][col]['set'] else ''
      else:
        label = translate_reference(result[row][col]['label'], globals)
      if 'condition' in result[row][col].keys():
        label = f"{label} [c]"
      labels[row].append(label)
  assert labels == [
    #['',           '0',            '1',            '2'], 
    ['',           'Epoch A',      'Epoch B',      'Epoch C'],
    ['',           'Screening',    'Dose',         'Check Up'], 
    ['',           '-2 Days',      'Dose',         '7 Days'], 
    ['',           '',             '',             '1..1 Days'], 
    ['Activity 1', 'X [c]',        '',             ''], 
    ['Activity 2', 'X',            'X',            ''], 
    ['Activity 3', '',             'X',            'X'], 
    ['Activity 4 [c]', '',         '',             'X'], 
    ['Activity 5', '',             '',             'X']
  ]




from usdm_excel.base_sheet import BaseSheet
from usdm_model.comment_annotation import CommentAnnotation
from usdm_excel.globals import Globals

class AnnotationSheet(BaseSheet):

  SHEET_NAME = 'notes'

  def __init__(self, file_path: str, globals: Globals):
    try:
      self.items = []
      super().__init__(file_path=file_path, globals=globals, sheet_name=self.SHEET_NAME, optional=True)
      if self.success:
        for index, row in self.sheet.iterrows():
          name = self.read_cell_by_name(index, 'name')
          text = self.read_cell_by_name(index, 'text')
          codes = self.read_other_code_cell_multiple_by_name(index, "codes")
          params = {'text': text, 'codes': codes}
          item = self.create_object(CommentAnnotation, params)
          if item:
            self.items.append(item)
            self.globals.cross_references.add(name, item)     
    except Exception as e:
      self._sheet_exception(e)



from typing import Union, List, Literal, Annotated
from annotated_types import Len
from .api_base_model import ApiBaseModelWithIdNameLabelAndDesc
from .code import Code
from .range import Range
from .characteristic import Characteristic
from .comment_annotation import CommentAnnotation

class PopulationDefinition(ApiBaseModelWithIdNameLabelAndDesc):
  includesHealthySubjects: bool
  plannedEnrollmentNumber: Union[Range, None] = None
  plannedCompletionNumber: Union[Range, None] = None
  plannedSex: Annotated[list[Code], Len(min_length=0, max_length=2)] = []
  criterionIds: List[str] = []
  plannedAge: Union[Range, None] = None
  notes: List[CommentAnnotation] = []
  instanceType: Literal['PopulationDefinition']

class StudyCohort(PopulationDefinition):
  characteristics: List[Characteristic] = []
  instanceType: Literal['StudyCohort']
  
class StudyDesignPopulation(PopulationDefinition):
  cohorts: List[StudyCohort] = []
  instanceType: Literal['StudyDesignPopulation']



from typing import List, Literal, Union
from .api_base_model import ApiBaseModelWithIdNameLabelAndDesc
from .schedule_timeline_exit import ScheduleTimelineExit
from .scheduled_instance import ScheduledActivityInstance, ScheduledDecisionInstance
from .timing import Timing

class ScheduleTimeline(ApiBaseModelWithIdNameLabelAndDesc):
  mainTimeline: bool
  entryCondition: str
  entryId: str
  exits: List[ScheduleTimelineExit] = []
  timings: List[Timing] = []
  instances: List[Union[ScheduledActivityInstance, ScheduledDecisionInstance]] = []
  instanceType: Literal['ScheduleTimeline']



import pytest
import pandas as pd
from usdm_excel.study_design_dictionary_sheet.study_design_dictionary_sheet import StudyDesignDictionarySheet
from usdm_model.api_base_model import ApiBaseModelWithId


xfail = pytest.mark.xfail

def test_create(mocker, globals):
  mock_cross_ref = mocker.patch("usdm_excel.cross_ref.CrossRef.get")
  mock_cross_ref.side_effect=[
    ApiBaseModelWithId(id="1"), ApiBaseModelWithId(id="2"), ApiBaseModelWithId(id="3"), ApiBaseModelWithId(id="4"), ApiBaseModelWithId(id="5")
  ]
  mock_present = mocker.patch("usdm_excel.base_sheet.BaseSheet._sheet_present")
  mock_present.side_effect=[True]
  mock_id = mocker.patch("usdm_excel.id_manager.IdManager.build_id")
  mock_id.side_effect=['DictionaryId_1', 'MapId_1', 'MapId_2', 'DictionaryId_2', 'MapId_3', 'DictionaryId_3', 'MapId_4', 'MapId_5', 'MapId_6']
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = [
    ['Dictionary 1', 'Dictionary One',   'Label One',   'Key 1', ApiBaseModelWithId, 'Id 1', 'Attribute 1', ''], 
    ['',             '',                 '',            'Key 2', ApiBaseModelWithId, 'Id 2', 'Attribute 2', ''], 
    ['Dictionary 2', 'Dictionary Two',   'Label Two',   'Key 3', ApiBaseModelWithId, 'Id 3', 'Attribute 3', ''], 
    ['Dictionary 3', 'Dictionary Three', 'Label Three', 'Key 4', ApiBaseModelWithId, 'Id 4', 'Attribute 4', ''], 
    ['',             '',                 '',            'Key 5', ApiBaseModelWithId, 'Id 5', 'Attribute 5', ''],
    ['',             '',                 '',            'Key 6', ''                , '',     '',            'Hello!']
 ]
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['name', 'description', 'label', 'key', 'class', 'xref', 'attribute', 'value'])
  dictionaries = StudyDesignDictionarySheet("", globals)
  assert len(dictionaries.items) == 3
  assert dictionaries.items[0].id == 'DictionaryId_1'
  assert dictionaries.items[0].name == 'Dictionary 1'
  assert dictionaries.items[0].description == 'Dictionary One'
  assert dictionaries.items[0].label == 'Label One'
  assert dictionaries.items[0].parameterMaps[1].tag == 'Key 2'
  assert dictionaries.items[0].parameterMaps[1].reference == '<usdm:ref klass="ApiBaseModelWithId" id="2" attribute="Attribute 2"></usdm:ref>'
  assert [x.tag for x in dictionaries.items[1].parameterMaps] == ['Key 3']
  assert [x.tag for x in dictionaries.items[2].parameterMaps] == ['Key 4', 'Key 5', 'Key 6']
  assert dictionaries.items[2].parameterMaps[2].reference == "Hello!"
  
def test_create_empty(mocker, globals):
  mock_present = mocker.patch("usdm_excel.base_sheet.BaseSheet._sheet_present")
  mock_present.side_effect=[True]
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = []
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['name', 'description', 'label', 'key', 'class', 'xref', 'attribute'])
  dictionaries = StudyDesignDictionarySheet("", globals)
  assert len(dictionaries.items) == 0

def test_read_cell_by_name_error(mocker, globals):
  
  call_parameters = []
  
  def my_add(sheet, row, column, message, level=10):
    call_parameters.append((sheet, row, column, message, level))
    return None

  globals.cross_references.clear()
  mock_present = mocker.patch("usdm_excel.base_sheet.BaseSheet._sheet_present")
  mock_present.side_effect=[True]
  mock_error = mocker.patch("usdm_excel.errors_and_logging.errors.Errors.add", side_effect=my_add)
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = [
    ['Dictionary 1', 'Dictionary One',   'Label One',   'Key 1', 'Klass 1', 'Attribute 1']
 ]
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['name', 'description', 'label', 'key', 'class', 'attribute'])
  dictionaries = StudyDesignDictionarySheet("", globals)
  mock_error.assert_called()
  assert call_parameters == [
    ("dictionaries", 1, -1, "Error attempting to read cell 'xref'. Exception: Failed to detect column(s) 'xref' in sheet", 40),
    ('dictionaries', 1, 6, "Failed to translate reference path 'Attribute 1', could not find start instance 'Klass 1', ''", 40)
  ]
  



from usdm_excel.globals import Globals
from usdm_excel.syntax_template_sheet import SyntaxTemplateSheet
from usdm_model.objective import Objective
from usdm_model.endpoint import Endpoint
from usdm_model.syntax_template_dictionary import SyntaxTemplateDictionary
from usdm_excel.globals import Globals

class StudyDesignObjectiveEndpointSheet(SyntaxTemplateSheet):

  SHEET_NAME = 'studyDesignOE'
  
  def __init__(self, file_path: str, globals: Globals):
    try:
      self.objectives = []
      super().__init__(file_path=file_path, globals=globals, sheet_name=self.SHEET_NAME)
      current = None
      for index, row in self.sheet.iterrows():
        o_text = self.read_cell_by_name(index, 'objectiveText') 
        ep_name = self.read_cell_by_name(index, ['endpointXref', 'endpointName'])
        ep_description = self.read_cell_by_name(index, 'endpointDescription')
        ep_label = self.read_cell_by_name(index, ["endpointLabel"], must_be_present=False) 
        ep_text = self.read_cell_by_name(index, 'endpointText') 
        ep_purpose = self.read_cell_by_name(index, ['endpointPurposeDescription', 'endpointPurpose'], default='None provided')
        ep_level = self.read_cdisc_klass_attribute_cell_by_name('Endpoint', 'endpointLevel', index, "endpointLevel")
        ep_dictionary_name = self.read_cell_by_name(index, 'endpointDictionary', must_be_present=False)
        self._validate_references(index, 'endpointText', ep_text, ep_dictionary_name)

        if o_text:
          o_name = self.read_cell_by_name(index, ["objectiveXref", "objectiveName"]) 
          o_description = self.read_cell_by_name(index, 'objectiveDescription')
          o_label = self.read_cell_by_name(index, ["objectiveLabel"], must_be_present=False) 
          o_level = self.read_cdisc_klass_attribute_cell_by_name('Objective', 'objectiveLevel', index, "objectiveLevel")
          o_dictionary_name = self.read_cell_by_name(index, 'objectiveDictionary', must_be_present=False)
          self._validate_references(index, 'objectiveText', o_text, o_dictionary_name)
          try:
            dictionary_id = self._get_dictionary_id(o_dictionary_name)
            current = Objective(id=self.globals.id_manager.build_id(Objective),
              name=o_name,
              description=o_description, 
              label=o_label,
              text=o_text,
              level=o_level,
              endpoints=[],
              dictionaryId=dictionary_id
            )
          except Exception as e:
            self._general_exception(f"Failed to create Objective object", e)
          else:
            self.objectives.append(current)
            self.globals.cross_references.add(o_name, current)
        if current is not None:
          try:
            dictionary_id = self._get_dictionary_id(ep_dictionary_name)
            ep = Endpoint(id=self.globals.id_manager.build_id(Endpoint),
              name=ep_name,
              description=ep_description,
              label=ep_label,
              text=ep_text, 
              purpose=ep_purpose, 
              level=ep_level,
              dictionaryId=dictionary_id
            )  
          except Exception as e:
            self._general_exception(f"Failed to create Endpoint object", e)
          else:
            current.endpoints.append(ep)
            self.globals.cross_references.add(ep_name, ep)
        else:
          self._general_error("Failed to add Endpoint, no Objective set")

    except Exception as e:
      self._sheet_exception(e)



import pandas as pd
from usdm_excel.study_design_sites_sheet.study_design_sites_sheet import StudyDesignSitesSheet
from usdm_model.code import Code


def test_create(mocker, globals):
  globals.cross_references.clear()
  mock_present = mocker.patch("usdm_excel.base_sheet.BaseSheet._sheet_present")
  mock_present.side_effect=[True]
  mock_id = mocker.patch("usdm_excel.id_manager.IdManager.build_id")
  mock_id.side_effect=['Site_1', 'Org_1', 'Addr_1', 'Id_1', 'Site_2', 'Org_2', 'Addr_2', 'Id_2', 'Site_3', 'Org_3', 'Addr_3', 'Id_3']
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = [
    [ 'CRO1', 'CRO One',   'Study Registry', 'DUNS', '111111111', 'line|district|city|state|postal_code|GBR',      'Site1', 'Site One', 'Big Site', 'GBR' ],
    [ 'CRO2', 'CRO Two',   'Study Registry', 'DUNS', '222222222', 'line2,district2,city2,state2,postal_code2,FRA', 'Site2', 'Site Two', 'Little Site', 'FRA'],
    [ 'CRO3', 'CRO Three', 'Study Registry', 'DUNS', '333333333', 'line3,district3,city3,state3,postal_code3,FR',  'Site3', 'Site Three', 'Middle Site', '']
  ]
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['name', 'label', 'type', 'identifierScheme', 'identifier', 'address', 'siteName', 'siteLabel', 'siteDescription', 'siteCountry'])
  mock_json = mocker.patch("json.load")
  mock_json.return_value = {}
  expected_1 = Code(id='Code1', code='code', codeSystem='codesys', codeSystemVersion='3', decode="GBR")
  expected_2 = Code(id='Code1', code='code', codeSystem='codesys', codeSystemVersion='3', decode="GBR")
  expected_3 = Code(id='Code2', code='code', codeSystem='codesys', codeSystemVersion='3', decode="FRA")
  expected_4 = Code(id='Code2', code='code', codeSystem='codesys', codeSystemVersion='3', decode="FRA")
  expected_5 = Code(id='Code3', code='code', codeSystem='codesys', codeSystemVersion='3', decode="FRA")
  expected_6 = Code(id='Code2', code='code', codeSystem='codesys', codeSystemVersion='3', decode="GER")
  mock_code = mocker.patch("usdm_excel.iso_3166.ISO3166.code")
  mock_code.side_effect=[expected_1, expected_2, expected_3, expected_4, expected_5, expected_6]
  sheet = StudyDesignSitesSheet("", globals)
  assert len(sheet.organizations) == 3
  assert len(sheet.sites) == 3
  assert sheet.organizations[0].model_dump() == {
    'id': 'Id_1',
    'identifier': '111111111',
    'identifierScheme': 'DUNS',
    'instanceType': 'Organization',
    'label': 'CRO One',
    'legalAddress': {
        'city': 'city',
        'country': {
            'code': 'code',
            'codeSystem': 'codesys',
            'codeSystemVersion': '3',
            'decode': 'GBR',
            'id': 'Code1',
            'instanceType': 'Code',
        },
        'district': 'district',
        'id': 'Addr_1',
        'instanceType': 'Address',
        'lines': [
            'line',
        ],
        'postalCode': 'postal_code',
        'state': 'state',
        'text': 'line, city, district, state, postal_code, GBR',
    },
    'managedSites': [
      {
        'country': {
          'code': 'code',
          'codeSystem': 'codesys',
          'codeSystemVersion': '3',
          'decode': 'GBR',
          'id': 'Code1',
          'instanceType': 'Code',
        },
        'description': 'Big Site',
        'id': 'Site_1',
        'instanceType': 'StudySite',
        'label': 'Site One',
        'name': 'Site1',
      },
    ],
    'name': 'CRO1',
    'type': {
      'code': 'C93453',
      'codeSystem': 'http://www.cdisc.org',
      'codeSystemVersion': '2023-12-15',
      'decode': 'Study Registry',
      'id': 'Org_1',
      'instanceType': 'Code',
    }
  }
  assert sheet.organizations[1].model_dump() == {
    'id': 'Id_2',
    'identifier': '222222222',
    'identifierScheme': 'DUNS',
    'instanceType': 'Organization',
    'label': 'CRO Two',
    'legalAddress': {
        'city': 'city2',
        'country': {
            'code': 'code',
            'codeSystem': 'codesys',
            'codeSystemVersion': '3',
            'decode': 'FRA',
            'id': 'Code2',
            'instanceType': 'Code',
        },
        'district': 'district2',
        'id': 'Addr_2',
        'instanceType': 'Address',
        'lines': [
            'line2',
        ],
        'postalCode': 'postal_code2',
        'state': 'state2',
        'text': 'line2, city2, district2, state2, postal_code2, FRA',
    },
    'managedSites': [
      {
        'country': {
          'code': 'code',
          'codeSystem': 'codesys',
          'codeSystemVersion': '3',
          'decode': 'FRA',
          'id': 'Code2',
          'instanceType': 'Code',
        },
        'description': 'Little Site',
        'id': 'Site_2',
        'instanceType': 'StudySite',
        'label': 'Site Two',
        'name': 'Site2',
      },
    ],
    'name': 'CRO2',
    'type': {
      'code': 'C93453',
      'codeSystem': 'http://www.cdisc.org',
      'codeSystemVersion': '2023-12-15',
      'decode': 'Study Registry',
      'id': 'Org_2',
      'instanceType': 'Code',
    }
  }
  
def test_create_empty(mocker, globals):
  globals.cross_references.clear()
  mock_present = mocker.patch("usdm_excel.base_sheet.BaseSheet._sheet_present")
  mock_present.side_effect=[True]
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = []
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['studyIdentifierName', 'studyIdentifierDescription', 'studyIdentifierType'])
  sites = StudyDesignSitesSheet("", globals)
  assert len(sites.organizations) == 0

def test_read_cell_by_name_error(mocker, globals):
  globals.cross_references.clear()
  mock_present = mocker.patch("usdm_excel.base_sheet.BaseSheet._sheet_present")
  mock_present.side_effect=[True]
  mock_error = mocker.patch("usdm_excel.errors_and_logging.errors.Errors.add")
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = [
    [ 'CRO1', 'CRO One', 'DUNS', '111111111', 'line|district|city|state|postal_code|GBR', 'Site1', 'Site One', 'Big Site', '' ],
  ]
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['name', 'label', 'identifierScheme', 'identifier', 'address', 'siteName', 'siteLabel', 'siteDescription', 'siteCountry'])
  mock_json = mocker.patch("json.load")
  mock_json.return_value = {}
  sites = StudyDesignSitesSheet("", globals)
  mock_error.assert_called()
  assert mock_error.call_args[0][0] == "studyDesignSites"
  assert mock_error.call_args[0][1] == None
  assert mock_error.call_args[0][2] == None
  assert mock_error.call_args[0][3] == "Exception. Error [Failed to detect column(s) 'type' in sheet] while reading sheet 'studyDesignSites'. See log for additional details."



import pandas as pd
from usdm_excel.study_identifiers_sheet.organization import get_organization
from usdm_excel.base_sheet import BaseSheet
from usdm_model.code import Code

def test_success(mocker, globals):
  ids = ['Code_1', 'Addr_1', 'Code_2', 'Org_1']
  expected = ( '{"id": "Org_1", "name": "ClinicalTrials.gov", "label": "", '
               '"type": {"id": "Code_1", "code": "C93453", "codeSystem": "http://www.cdisc.org", "codeSystemVersion": "2023-12-15", "decode": "Study Registry", "instanceType": "Code"}, '
               '"identifierScheme": "USGOV", "identifier": "CT-GOV", '
               '"legalAddress": {"id": "Addr_1", "text": "line, city, district, state, postal_code, Denmark", "lines": ["line"], "city": "city", "district": "district", "state": "state", "postalCode": "postal_code", '
                 '"country": {"id": "Code_2", "code": "DNK", "codeSystem": "ISO 3166 1 alpha3", "codeSystemVersion": "2020-08", "decode": "Denmark", "instanceType": "Code"}, '
                 '"instanceType": "Address"}, '
               '"managedSites": [], '
               '"instanceType": "Organization"}'
  )
  data = {'organisationIdentifierScheme': ['USGOV'], 'organisationIdentifier': ['CT-GOV'], 'organisationName': ['ClinicalTrials.gov'], 'organisationType': ['Study Registry'], 'studyIdentifier': ['NCT12345678'], 'organisationAddress': ['line|district|city|state|postal_code|GBR']}
  base = _setup_base(mocker, globals, data, ids)
  item = get_organization(base, 0)
  assert str(item.to_json()) == expected

def test_no_address(mocker, globals):
  ids = ['Code_1', 'Org_1']
  expected = ( '{"id": "Org_1", "name": "ClinicalTrials.gov", "label": "", '
               '"type": {"id": "Code_1", "code": "C93453", "codeSystem": "http://www.cdisc.org", "codeSystemVersion": "2023-12-15", "decode": "Study Registry", "instanceType": "Code"}, '
               '"identifierScheme": "USGOV", "identifier": "CT-GOV", '
               '"legalAddress": null, '
               '"managedSites": [], '
               '"instanceType": "Organization"}'
  )
  data = {'organisationIdentifierScheme': ['USGOV'], 'organisationIdentifier': ['CT-GOV'], 'organisationName': ['ClinicalTrials.gov'], 'organisationType': ['Study Registry'], 'studyIdentifier': ['NCT12345678'], 'organisationAddress': ['']}
  base = _setup_base(mocker, globals, data, ids)
  mock_error = mocker.patch("usdm_excel.errors_and_logging.errors.Errors.add")
  item = get_organization(base, 0)
  assert str(item.to_json()) == expected
  assert mock_error.call_count == 1
  mock_error.assert_has_calls([mocker.call('sheet', 1, 6, "Address '' does not contain the required fields (lines, district, city, state, postal code and country code) using ',' separator characters, only 0 found", 40)])

def test_organization_error(mocker, globals):
  ids = ['Code_1', 'Addr_1', 'Code_2', 'Org_1']
  data = {'organisationIdentifierScheme': ['USGOV'], 'organisationIdentifier': ['CT-GOV'], 'organisationName': [''], 'organisationType': ['Study Registry'], 'studyIdentifier': ['NCT12345678'], 'organisationAddress': ['line|district|city|state|postal_code|GBR']}
  base = _setup_base(mocker, globals, data, ids)
  mock_error = mocker.patch("usdm_excel.errors_and_logging.errors.Errors.add")
  item = get_organization(base, 0)
  assert item is None
  assert mock_error.call_count == 1
  mock_error.assert_has_calls([mocker.call('sheet', None, None, 'Exception. Failed to create Organization object. See log for additional details.', 40)])

def _setup_base(mocker, globals, data, ids):
  globals.cross_references.clear()
  mock_id = mocker.patch("usdm_excel.id_manager.IdManager.build_id")
  mock_id.side_effect=ids
  mock_json = mocker.patch("json.load")
  mock_json.return_value = {}
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame.from_dict(data)
  base = BaseSheet("", globals, "sheet")
  return base


from usdm_excel.study_soa_v2_sheet.soa_column_rows import SoAColumnRows
from usdm_excel.study_soa_v2_sheet.scheduled_instance import ScheduledInstance
#from usdm_excel.cross_ref import cross_references
#from usdm_excel.id_manager import id_manager
from usdm_model.schedule_timeline_exit import ScheduleTimelineExit
from usdm_model.scheduled_instance import ConditionAssignment

class ScheduledInstances():
  
  def __init__(self, parent):
    self.parent = parent
    self.items = []
    self.map = {}
    self.exits = []
    self.instances = []
    self._build_instances()
    self._set_default_references()
    self._set_condition_references()

  def match(self, name):
    return self.map[name] if name in self.map else None
  
  def _build_instances(self):    
    for col_index in range(self.parent.sheet.shape[1]):
      if col_index >= SoAColumnRows.FIRST_VISIT_COL:
        record = ScheduledInstance(self.parent, col_index)
        self.items.append(record)
        self.map[record.name] = record

  def _set_default_references(self):
    for instance in self.items:
      item = instance.item
      self.instances.append(item)
      if instance.default_name in self.map.keys():
        instance.item.defaultConditionId = self.map[instance.default_name].item.id
      elif instance.default_name.upper() == "(EXIT)":
        exit = self._add_exit()
        item.timelineExitId = exit.id
        self.exits.append(exit)
      else:
        self.parent._general_error(f"Default reference from {instance.name} to {instance.default_name} cannot be made, not found on the same timeline")

  def _add_exit(self):
    exit = ScheduleTimelineExit(id=self.parent.globals.id_manager.build_id(ScheduleTimelineExit))
    self.parent.globals.cross_references.add(exit.id, exit)
    return exit

  def _set_condition_references(self):
    for instance in self.items:
      item = instance.item
      if item.instanceType == 'ScheduledDecisionInstance':
        for condition in instance.conditions.items:
          #print(f"COND: {condition} ")
          if condition['name'] in self.map.keys():
            ca = self.parent.create_object(ConditionAssignment, {'condition': condition['condition'], 'conditionTargetId': self.map[condition['name']].item.id})
            if ca:
              item.conditionAssignments.append(ca)
          else:
            self.parent._general_error(f"Conditonal reference from {instance.name} to {condition['name']} cannot be made, not found on the same timeline")
  


import json
import csv
import yaml
from usdm_db import USDMDb
from bs4 import BeautifulSoup
from uuid import UUID

SAVE_ALL = False

def save_error_csv(file, contents):
  writer = csv.DictWriter(file, fieldnames=['sheet','row','column','message','level'])
  writer.writeheader()
  writer.writerows(contents)

def to_int(value):
  try:
    return int(value)
  except:
    return None

def read_error_csv(file):
  reader = csv.DictReader(file)
  items = list(reader)
  for item in items:
    item['row'] = to_int(item['row'])
    item['column'] = to_int(item['column'])
  return items

def prep_errors_for_csv_compare(errors):
  for error in errors:
    error['sheet'] = error['sheet'] if error['sheet'] else ''
  return errors

def format_html(result):
  soup = BeautifulSoup(result, 'html.parser')
  return soup.prettify()

def run_test(filename, save=False):
  usdm = USDMDb()
  errors = usdm.from_excel(f"tests/integration_test_files/{filename}.xlsx")
  result = usdm.to_json()
  #errors = excel.errors()

  # Useful if you want to see the results.
  if save or SAVE_ALL:
    with open(f"tests/integration_test_files/{filename}.json", 'w', encoding='utf-8') as f:
      f.write(json.dumps(json.loads(result), indent=2))
    with open(f"tests/integration_test_files/{filename}_errors.csv", 'w',newline='') as f:
      save_error_csv(f, errors) 
  
  with open(f"tests/integration_test_files/{filename}.json", 'r') as f:
    expected = json.dumps(json.load(f)) # Odd, but doing it for consistency of processing
  assert result == expected
  with open(f"tests/integration_test_files/{filename}_errors.csv", 'r') as f:
    expected = read_error_csv(f)
  assert prep_errors_for_csv_compare(errors) == expected

def run_test_html(filename, save=False, highlight=False):
  suffix = "_highlight" if highlight else ""
  usdm = USDMDb()
  errors = usdm.from_excel(f"tests/integration_test_files/{filename}.xlsx")
  result = usdm.to_html("SPONSOR", highlight)

  # Useful if you want to see the results.
  if save or SAVE_ALL:
    with open(f"tests/integration_test_files/{filename}{suffix}.html", 'w') as f:
      f.write(format_html(result))
    with open(f"tests/integration_test_files/{filename}{suffix}_html_errors.csv", 'w',newline='') as f:
      save_error_csv(f, errors) 

  with open(f"tests/integration_test_files/{filename}{suffix}.html", 'r') as f:
    expected = f.read()
  assert format_html(result) == expected
  with open(f"tests/integration_test_files/{filename}{suffix}_html_errors.csv", 'r') as f:
    expected = read_error_csv(f)
  assert prep_errors_for_csv_compare(errors) == expected

def run_test_timeline(filename, level=USDMDb.FULL_HTML, save=False):
  usdm = USDMDb()
  usdm.from_excel(f"tests/integration_test_files/{filename}.xlsx")
  result = usdm.to_timeline(level)

  # Useful if you want to see the results.
  if save or SAVE_ALL:
    with open(f"tests/integration_test_files/{filename}_timeline_{level}.html", 'w') as f:
      f.write(result)
  
  with open(f"tests/integration_test_files/{filename}_timeline_{level}.html", 'r') as f:
    expected = f.read()
  assert result == expected

def run_test_neo4j(filename, mocker, save=False):
  fake_uuids = (UUID(f'00000000-0000-4000-8000-{i:012}', version=4) for i in range(10000))
  mocker.patch("usdm_db.neo4j_dict.uuid4", side_effect=fake_uuids)
  usdm = USDMDb()
  usdm.from_excel(f"tests/integration_test_files/{filename}.xlsx")
  result = usdm.to_neo4j_dict()

  # Useful if you want to see the results.
  if save or SAVE_ALL:
    with open(f"tests/integration_test_files/{filename}_neo4j_dict.yaml", 'w') as f:
      f.write(yaml.dump(result))
  
  with open(f"tests/integration_test_files/{filename}_neo4j_dict.yaml", 'r') as f:
    expected = yaml.safe_load(f) 
  assert result == expected

def run_test_fhir(filename, mocker, save=False):
  fake_uuids = (UUID(f'00000000-0000-4000-8000-{i:012}', version=4) for i in range(1,10))
  mocker.patch("usdm_db.uuid4", side_effect=fake_uuids)
  usdm = USDMDb()
  usdm.from_excel(f"tests/integration_test_files/{filename}.xlsx")
  result = usdm.to_fhir('sponsor')

  print(f"RESULT: {result}")

  if save or SAVE_ALL:
    with open(f"tests/integration_test_files/{filename}_fhir.json", 'w', encoding='utf-8') as f:
      f.write(json.dumps(json.loads(result), indent=2))
  
  with open(f"tests/integration_test_files/{filename}_fhir.json", 'r') as f:
    expected = json.load(f)
  result_dict = json.loads(result)
  result_dict["entry"][0]["resource"]["date"] = expected["entry"][0]["resource"]["date"] # Date is dynamic, bit of a fiddle but datetime mocking is a pain.
  json.dumps(result_dict) == json.dumps(expected)

def test_full_1():
  run_test('full_1')

def test_full_1_timeline():
  run_test_timeline('full_1')

def test_full_1_timeline_body():
  run_test_timeline('full_1', USDMDb.BODY_HTML)

def test_full_1_html():
  run_test_html('full_1')

def test_full_2():
  run_test('full_2')

def test_full_2_html():
  run_test_html('full_2')

def test_full_3():
  run_test('full_3')

def test_full_3_html():
  run_test_html('full_3')

def test_full_4():
  run_test('full_4')

def test_full_5():
  run_test('full_5')

def test_full_6():
  run_test('full_6')

def test_encounter_1():
  run_test('encounter_1')

def test_simple_1():
  run_test('simple_1')

def test_scope_1():
  run_test('scope_1')

def test_amendment_1():
  run_test('amendment_1')

def test_eligibility_criteria_1():
  run_test('eligibility_criteria_1')

def test_simple_1_html():
  run_test_html('simple_1')

def test_config_1():
  run_test('config_1')

def test_config_2():
  run_test('config_2')

def test_no_activity_sheet():
  run_test('no_activity_sheet')

def test_address_comma():
  run_test('address')

def test_complex_1():
  run_test('complex_1')

def test_arms_epochs_1():
  run_test('arms_epochs')

def test_cycles_1():
  run_test('cycles_1')

def test_multiple_column_names():
  run_test('multiple_column_names')

def test_invalid_section_levels():
  run_test('invalid_section_levels')

def test_new_field_names():
  run_test('new_labels')

def test_timing_check():
  run_test('timing_check')

def test_v2_soa_1():
  run_test('simple_1_v2')

def test_v2_soa_2():
  run_test('decision_1_v2')

def test_timeline_1():
  run_test_timeline('simple_1_v2')

def test_timeline_2():
  run_test_timeline('decision_1_v2')

def test_timeline_3():
  run_test_timeline('complex_1')

def test_timeline_4():
  run_test_timeline('cycles_1')

def test_path_error():
  run_test('path_error')

def test_references():
  run_test('references')

def test_references_html():
  run_test_html('references')

def test_references_html_highlight():
  run_test_html('references', highlight=True)

def test_simple_neo4j_1(mocker):
  run_test_neo4j('simple_1', mocker)

def test_full_neo4j_1(mocker):
  run_test_neo4j('full_1', mocker)

def test_full_neo4j_2(mocker):
  run_test_neo4j('full_2', mocker)

def test_full_neo4j_3(mocker):
  run_test_neo4j('full_3', mocker)

def test_full_fhir_1(mocker):
  run_test_fhir('full_1', mocker)

def test_group_activities_1(mocker):
  run_test('group_activities_1', mocker)

def test_notes_1(mocker):
  run_test('annotations_and_abbreviations', mocker)



from usdm_excel.base_sheet import BaseSheet
from usdm_excel.iso_3166 import ISO3166
from usdm_model.study_site import StudySite
from usdm_model.organization import Organization
from usdm_excel.globals import Globals

class StudyDesignSitesSheet(BaseSheet):

  SHEET_NAME = 'studyDesignSites'
  
  def __init__(self, file_path: str, globals: Globals):
    try:
      self.organizations: list[Organization] = []
      self.sites: list[StudySite] = []
      super().__init__(file_path=file_path, globals=globals, sheet_name=self.SHEET_NAME, optional=True)
      if self.success:
        current_org: Organization = None
        for index, row in self.sheet.iterrows():
          org_name = self.read_cell_by_name(index, 'name')
          site_name = self.read_cell_by_name(index, 'siteName')
          site_description = self.read_cell_by_name(index, 'siteDescription')
          site_label = self.read_cell_by_name(index, 'siteLabel')
          site_country = ISO3166(self.globals).code(self.read_cell_by_name(index, 'siteCountry'))
          site = self.create_object(StudySite, {'name': site_name, 'description': site_description, 'label': site_label, 'country': site_country})
          if site:
            self.sites.append(site)
            self.globals.cross_references.add(site.name, site)     
          if org_name:
            org_label = self.read_cell_by_name(index, 'label')
            org_type = self.read_cdisc_klass_attribute_cell_by_name('Organization', 'organizationType', index, ['type'])     
            org_id_scheme = self.read_cell_by_name(index, 'identifierScheme')
            org_identifier = self.read_cell_by_name(index, 'identifier')
            org_address = self.read_address_cell_by_name(index, 'address')
            if org_address:
              self.globals.cross_references.add(org_address.id, org_address)   
            item = self.create_object(Organization, {'identifierScheme': org_id_scheme, 'identifier': org_identifier, 'name': org_name, 'label': org_label, 'type': org_type, 'legalAddress': org_address, 'managedSites': [site]})
            if item:
              self.organizations.append(item)
              self.globals.cross_references.add(item.name, item)     
              current_org = item
          else:
            current_org.managedSites.append(site)
    except Exception as e:
      self._sheet_exception(e)

  def _site_country(self, index: int):
    return 


from typing import Literal
from .api_base_model import ApiBaseModelWithIdAndName
from .quantity import Quantity
from .code import Code

class AdministrableProductProperty(ApiBaseModelWithIdAndName):
  text: str
  type: Code
  quantity: Quantity
  instanceType: Literal['AdministrableProductProperty']



import traceback
from usdm_excel.base_sheet import BaseSheet
from usdm_model.indication import Indication
from usdm_excel.globals import Globals

class StudyDesignIndicationSheet(BaseSheet):

  SHEET_NAME = 'studyDesignIndications'
  
  def __init__(self, file_path: str, globals: Globals):
    try:
      self.items = []
      super().__init__(file_path=file_path, globals=globals, sheet_name=self.SHEET_NAME )
      for index, row in self.sheet.iterrows():
        name = self.read_cell_by_name(index, "name")
        description = self.read_cell_by_name(index, "description")
        label = self.read_cell_by_name(index, 'label', default="")
        rare = self.read_boolean_cell_by_name(index, 'isRareDisease', must_be_present=False)
        codes = self.read_other_code_cell_multiple_by_name(index, "codes")
        item = self.create_object(Indication, {'name': name, 'description': description, 'label': label, 'isRareDisease': rare, 'codes': codes})
        if item:
          self.items.append(item)
          self.globals.cross_references.add(name, item)
    except Exception as e:
      self._sheet_exception(e)




from typing import Literal, Union
from .api_base_model import ApiBaseModelWithIdNameLabelAndDesc
from .quantity import Quantity
from .geographic_scope import GeographicScope

class SubjectEnrollment(ApiBaseModelWithIdNameLabelAndDesc):
  quantity: Quantity
  appliesTo: Union[GeographicScope, None] = None
  appliesToId: Union[str, None] = None
  instanceType: Literal['SubjectEnrollment']



from typing import List, Literal, Union
from .api_base_model import ApiBaseModelWithIdAndName

class NarrativeContentItem(ApiBaseModelWithIdAndName):
  text: str
  instanceType: Literal['NarrativeContentItem']

class NarrativeContent(ApiBaseModelWithIdAndName):
  sectionNumber: Union[str, None] = None
  sectionTitle: Union[str, None] = None
  displaySectionNumber: bool
  displaySectionTitle: bool
  childIds: List[str] = []
  previousId: Union[str, None] = None
  nextId: Union[str, None] = None
  contentItemId: Union[str, None] = None
  instanceType: Literal['NarrativeContent']



import pandas as pd
from usdm_excel.base_sheet import BaseSheet
from usdm_model.code import Code
from usdm_model.alias_code import AliasCode

class Factory():

  def __init__(self, globals):
    self.globals = globals

  def item(self, cls, params):
    params['id'] = params['id'] if 'id' in params else self.globals.id_manager.build_id(cls)
    params['instanceType'] = cls.__name__
    return cls(**params)

  def set(self, cls, item_list):
    results = []
    for item in item_list:
      results.append(self.item(cls, item))
    return results

  def base_sheet(self, mocker):
    mocked_open = mocker.mock_open(read_data="File")
    mocker.patch("builtins.open", mocked_open)
    data = []
    mock_read = mocker.patch("pandas.read_excel")
    mock_read.return_value = pd.DataFrame(data, columns=[])
    return BaseSheet("", self.globals, "")

  def code(self, code, decode):
    return self._build_code(code=code, system="yyy", version="2", decode=decode)

  def geo_code(self, code, decode):
    return self._build_code(code=code, system="ISO 3166 1 alpha3", version="2020-08", decode=decode)

  def cdisc_code(self, code, decode):
    return self._build_code(code=code, system="xxx", version="1", decode=decode)

  def english(self):
    return self._build_code(code="en", system="ISO639", version="2007", decode="English")

  def cdisc_dummy(self):
    return self.cdisc_code("C12345", "decode")

  def alias_code(self, standard_code, alias_codes=[]):
    return self.item(AliasCode, {'standardCode': standard_code, 'standardCodeAliases': alias_codes})

  def _build_code(self, code, system, version, decode):
    return self.item(Code, {'code': code, 'codeSystem': system, 'codeSystemVersion': version, 'decode': decode})



from typing import Union, List, Literal
from .api_base_model import ApiBaseModelWithIdNameLabelAndDesc
from .transition_rule import TransitionRule
from .comment_annotation import CommentAnnotation

class StudyElement(ApiBaseModelWithIdNameLabelAndDesc):
  transitionStartRule: Union[TransitionRule, None] = None
  transitionEndRule: Union[TransitionRule, None] = None
  studyInterventionIds: List[str] = []
  notes: List[CommentAnnotation] = []
  instanceType: Literal['StudyElement']



from .utility import usdm_reference
from usdm_excel.base_sheet import BaseSheet
from usdm_model.schedule_timeline import ScheduleTimeline
#from usdm_model.activity import Activity
#from usdm_model.scheduled_instance import ScheduledActivityInstance, ScheduledDecisionInstance
from usdm_model.study_design import StudyDesign

class SoA():

  def __init__(self, parent: BaseSheet, study_design: StudyDesign, timeline: ScheduleTimeline):
    self.parent = parent
    self.study_design = study_design
    self.timeline = timeline

  def _first(self, collection):
    return next((item for item in collection if not item.previousId), None)

  def _find(self, collection, id):
    return next((item for item in collection if item.id == id), None)

  def _timing_from(self, timings, sai):
    return next((item for item in timings if item.relativeFromScheduledInstanceId == sai.id), None)

  def _condition_no_context(self, conditions, target):
    return next((item for item in conditions if target.id in item.appliesToIds and not item.contextIds), None)

  def _condition_with_context(self, conditions, target, context):
    return next((item for item in conditions if target.id in item.appliesToIds and context.id in item.contextIds), None)

  def _template_copy(self, template):
    result = []
    for item in template:
      result.append(item.copy())
    return result

  def _has_non_empty(self, the_list):
    return any(s['label'] != '' for s in the_list)

  def generate(self):

    # ScheduleActivityInstance order
    required_activities = []
    sai_order = []
    item = self._find(self.timeline.instances, self.timeline.entryId)
    more = True
    while more:
      sai_order.append(item)
      if hasattr(item, 'activityIds'):
        for id in item.activityIds:
          if id not in required_activities:
            required_activities.append(id)
      more = False if not item.defaultConditionId else True
      item = self._find(self.timeline.instances, item.defaultConditionId)
    #print(f"SAI ORDER: {sai_order}\n\n")
    #print(f"ACTIVITIES: {required_activities}\n\n")

    # Activity order
    activity_order = []
    item = self._first(self.study_design.activities)
    more = True
    while more:
      #print(f"ACTIVITY CHECK: {item.id}\n\n")
      if item.id in required_activities:
        activity_order.append(item)
      more = False if not item.nextId else True
      item = self._find(self.study_design.activities, item.nextId)
    #print(f"ACTIVITY ORDER: {activity_order}\n\n")

    row_template = []
    lh_columns = ['activity']
    sai_start_index = len(lh_columns)
    for item in lh_columns:
      row_template.append({'label': ''})
    for sai in enumerate(sai_order):
      row_template.append({'label': ''})

    results = []

    # row = self._template_copy(row_template)
    # for index, sai in enumerate(sai_order):
    #   row[index + sai_start_index]['label'] = index
    # results.append(row)

    row = self._template_copy(row_template)    
    for index, sai in enumerate(sai_order):
      if sai.epochId:
        item = self.parent.globals.cross_references.get_by_id("StudyEpoch", sai.epochId)
        row[index + sai_start_index]['label'] = usdm_reference(item, 'label') if item else "???"
    if self._has_non_empty(row):
      results.append(row)

    row = self._template_copy(row_template)
    for index, sai in enumerate(sai_order):
      if hasattr(sai, 'encounterId'):
        if sai.encounterId:
          item = self.parent.globals.cross_references.get_by_id("Encounter", sai.encounterId)
          row[index + sai_start_index]['label'] = usdm_reference(item, 'label') if item else "???"
    if self._has_non_empty(row):
      results.append(row)

    row = self._template_copy(row_template)
    for index, sai in enumerate(sai_order):
      timing = self._timing_from(self.timeline.timings, sai)
      if timing:
        row[index + sai_start_index]['label'] = usdm_reference(timing, 'label')
    if self._has_non_empty(row):
      results.append(row)

    row = self._template_copy(row_template)
    for index, sai in enumerate(sai_order):
      timing = self._timing_from(self.timeline.timings, sai)
      if timing:
        row[index + sai_start_index]['label'] = usdm_reference(timing, 'windowLabel')
    if self._has_non_empty(row):
      results.append(row)

    for activity in activity_order:
      row = self._template_copy(row_template)
      row[0]['label'] = usdm_reference(activity, 'label')
      condition = self._condition_no_context(self.study_design.conditions, activity)
      if condition:
        row[0]['condition'] = condition
      for index, sai in enumerate(sai_order):
        row[index + sai_start_index]['set'] = False
        if hasattr(sai, 'activityIds'):
          if activity.id in sai.activityIds:
            row[index + sai_start_index]['set'] = True
            condition = self._condition_with_context(self.study_design.conditions, activity, sai)
            if condition:
              row[index + sai_start_index]['condition'] = condition
      results.append(row)

    return results


class SoAColumnRows:
  
  NAME_ROW = 0
  DESCRIPTION_ROW = 1
  LABEL_ROW = 2
  TYPE_ROW = 3
  DEFAULT_ROW = 4
  CONDITIONS_ROW = 5
  EPOCH_ROW = 6
  ENCOUNTER_ROW = 7
  
  HEADER_ROW = 8
  FIRST_ACTIVITY_ROW = 9

  ACTIVITY_COL = 0
  CHILD_ACTIVITY_COL = 1
  BC_COL = 2
  FIRST_VISIT_COL = 3



from typing import Literal
from .api_base_model import ApiBaseModelWithId

class DocumentContentReference(ApiBaseModelWithId):
  sectionNumber: str
  sectionTitle: str
  instanceType: Literal['DocumentContentReference']



from typing import List, Literal
from pydantic import Field
from .api_base_model import ApiBaseModelWithId
from .comment_annotation import CommentAnnotation

class Abbreviation(ApiBaseModelWithId):
  abbreviatedText: str = Field(min_length=1)
  expandedText: str = Field(min_length=1)
  notes: List[CommentAnnotation] = []
  instanceType: Literal['Abbreviation']



from typing import List, Union, Literal
from .api_base_model import ApiBaseModelWithIdNameLabelAndDesc
from .administrable_product_property import AdministrableProductProperty
from .administrable_product_identifier import AdministrableProductIdentifier
from .ingredient import Ingredient
from .code import Code
from .alias_code import AliasCode
from .comment_annotation import CommentAnnotation

class AdministrableProduct(ApiBaseModelWithIdNameLabelAndDesc):
  pharmacologicClass:	Union[Code, None] = None
  administrableDoseForm: AliasCode
  properties: List[AdministrableProductProperty] = []
  identifiers: List[AdministrableProductIdentifier] = []
  ingredients: List[Ingredient] = []
  notes: List[CommentAnnotation] = []
  instanceType: Literal['AdministrableProduct']



from typing import Literal
from .api_base_model import ApiBaseModelWithId
from .substance import Substance
from .code import Code

class Ingredient(ApiBaseModelWithId):
  role: Code
  substance: Substance
  instanceType: Literal['Ingredient']



from pydantic import BaseModel, constr
from typing import Union
import json
import enum
import datetime
from uuid import UUID

# Example, see https://stackoverflow.com/questions/10252010/serializing-class-instance-to-json
def _serialize_as_json(obj):
  if isinstance(obj, enum.Enum):
    return obj.value
  elif isinstance(obj, datetime.date):
    return obj.isoformat()
  elif isinstance(obj, UUID):
    return str(obj)
  else:
    return obj.__dict__

def _serialize_as_json_with_type(obj):
  if isinstance(obj, enum.Enum):
    return obj.value
  elif isinstance(obj, datetime.date):
    return obj.isoformat()
  elif isinstance(obj, UUID):
    return str(obj)
  else:
    result = obj.__dict__
    result['_type'] = obj.__class__.__name__
    return result

class ApiBaseModel(BaseModel):

  def __init__(self, *args, **kwargs):
    kwargs['instanceType'] = self.__class__.__name__
    super().__init__(*args, **kwargs)

  def to_json(self):
    return json.dumps(self, default=_serialize_as_json)

  def to_json_with_type(self):
    return json.dumps(self, default=_serialize_as_json_with_type)

class ApiBaseModelWithId(ApiBaseModel):
  id: constr(min_length=1)

class ApiBaseModelWithIdAndDesc(ApiBaseModelWithId):
  description: Union[str, None] = None

class ApiBaseModelWithIdAndName(ApiBaseModelWithId):
  name: constr(min_length=1)

class ApiBaseModelWithIdNameAndLabel(ApiBaseModelWithIdAndName):
  label: Union[str, None] = None

class ApiBaseModelWithIdNameLabelAndDesc(ApiBaseModelWithIdNameAndLabel):
  description: Union[str, None] = None

class ApiBaseModelWithIdNameAndDesc(ApiBaseModelWithIdAndName):
  description: Union[str, None] = None



import pytest
from usdm_excel.cross_ref import *

class CRTest():

  def __init__(self, id, name):
    self.id = id
    self.name = name
    self.value = "VALUE"

class CRTest2():

  def __init__(self, id, name, instance):
    self.id = id
    self.name = name
    self.child = instance
    self.value = "VALUE"

class CRTest3():

  def __init__(self, id, name, instance):
    self.id = id
    self.name = name
    self.child = instance
    self.value = "VALUE"

def test_create(globals):
  object = CrossRef(globals.errors_and_logging)
  assert len(object._references.keys()) == 0
  assert object._references == {}

def test_clear(globals):
  cross_references = globals.cross_references
  item = CRTest(id="1234", name="name")
  cross_references._references = {}
  cross_references._identifiers = {}
  cross_references._references["CRTest.name"] = item
  cross_references._identifiers["CRTest.1234"] = item
  assert len(cross_references._references.keys()) == 1
  assert len(cross_references._identifiers.keys()) == 1
  cross_references.clear()
  assert len(cross_references._references.keys()) == 0
  assert len(cross_references._identifiers.keys()) == 0

def test_add(globals):
  cross_references = globals.cross_references
  item = CRTest(id="1234", name="name")
  cross_references.clear()
  assert len(cross_references._references.keys()) == 0
  assert len(cross_references._identifiers.keys()) == 0
  cross_references.add("name", item)
  assert len(cross_references._references.keys()) == 1
  assert cross_references._references["CRTest.name"] == item
  assert len(cross_references._identifiers.keys()) == 1
  assert cross_references._identifiers["CRTest.1234"] == item

def test_get(globals):
  cross_references = globals.cross_references
  item = CRTest(id="1234", name="name")
  cross_references.clear()
  assert len(cross_references._references.keys()) == 0
  assert len(cross_references._identifiers.keys()) == 0
  cross_references._references["CRTest.name"] = item
  assert cross_references.get(CRTest, "name") == item

def test_get_by_id(globals):
  cross_references = globals.cross_references
  item = CRTest(id="1234", name="name")
  cross_references.clear()
  assert len(cross_references._references.keys()) == 0
  assert len(cross_references._identifiers.keys()) == 0
  cross_references._identifiers["CRTest.1234"] = item
  assert cross_references.get_by_id(CRTest, "1234") == item
  assert cross_references.get_by_id("CRTest", "1234") == item

def test_get_by_path(globals):
  cross_references = globals.cross_references
  item1 = CRTest(id="1234", name="name1")
  item2 = CRTest2(id="1235", name="name2", instance=item1)
  item3 = CRTest3(id="1236", name="name3", instance=item2)
  cross_references.clear()
  assert len(cross_references._references.keys()) == 0
  assert len(cross_references._identifiers.keys()) == 0
  cross_references.add("name1", item1)
  cross_references.add("name2", item2)  
  cross_references.add("name3", item3)
  instance, attribute = cross_references.get_by_path("CRTest3", "name3", "child/CRTest2/@child/CRTest/@value")
  assert instance.id == "1234"
  assert attribute == "value"
  instance, attribute = cross_references.get_by_path("CRTest3", "name3", "child/CRTest2/child/CRTest/@value")
  assert instance.id == "1234"
  assert attribute == "value"

def test_get_by_path_errors(globals):
  cross_references = globals.cross_references
  item1 = CRTest(id="1234", name="name1")
  item2 = CRTest2(id="1235", name="name2", instance=item1)
  item3 = CRTest3(id="1236", name="name3", instance=item2)
  cross_references.clear()
  assert len(cross_references._references.keys()) == 0
  assert len(cross_references._identifiers.keys()) == 0
  cross_references.add("name1", item1)
  cross_references.add("name2", item2)  
  cross_references.add("name3", item3)
  with pytest.raises(cross_references.PathError) as ex_info:
    instance, attribute = cross_references.get_by_path("CRTest4", "name3", "child/CRTest2/@child/CRTest/@value")
  assert str(ex_info.value) == "Failed to translate reference path 'child/CRTest2/@child/CRTest/@value', could not find start instance 'CRTest4', 'name3'"

  with pytest.raises(cross_references.PathError) as ex_info:
    instance, attribute = cross_references.get_by_path("CRTest3", "name3", "child/CRTest4/@child/CRTest/@value")
  assert str(ex_info.value) == "Failed to translate reference path 'child/CRTest4/@child/CRTest/@value', class mismtach, expecting 'CRTest4', found 'CRTest2'"

  with pytest.raises(cross_references.PathError) as ex_info:
    instance, attribute = cross_references.get_by_path("CRTest3", "name3", "child/CRTest2/@childXXX/CRTest/@value")
  assert str(ex_info.value) == "Failed to translate reference path 'child/CRTest2/@childXXX/CRTest/@value', attribute 'childXXX' was not found"

  with pytest.raises(cross_references.PathError) as ex_info:
    instance, attribute = cross_references.get_by_path("CRTest3", "name3", "child/CRTest2//CRTest/@value")
  assert str(ex_info.value) == "Failed to translate reference path 'child/CRTest2//CRTest/@value', attribute '' was not found"

  with pytest.raises(cross_references.PathError) as ex_info:
    instance, attribute = cross_references.get_by_path("CRTest3", "name3", "child/CRTest2/child/CRTest")
  assert str(ex_info.value) == "Failed to translate reference path 'child/CRTest2/child/CRTest', format error"



from usdm_excel.base_sheet import BaseSheet
from usdm_model.study_role import StudyRole
from usdm_model.assigned_person import AssignedPerson
from usdm_model.organization import Organization
from usdm_model.masking import Masking
from usdm_excel.globals import Globals

class StudyRoleSheet(BaseSheet):

  SHEET_NAME = 'roles'

  def __init__(self, file_path: str, globals: Globals):
    try:
      self.items = []
      super().__init__(file_path=file_path, globals=globals, sheet_name=self.SHEET_NAME, optional=True)
      if self.success:
        for index, row in self.sheet.iterrows():
          params = {
            'name': self.read_cell_by_name(index, 'name'), 
            'description': self.read_cell_by_name(index, 'description', default=''), 
            'label': self.read_cell_by_name(index, 'label', default=''), 
            'organizationIds': self._get_refs_for(Organization, index, 'organizations', ids=True), 
            'assignedPersons': self._get_refs_for(AssignedPerson, index, 'people'), 
            'masking': self._get_masking(index), 
            'code': self.read_cdisc_klass_attribute_cell_by_name("StudyRole", "code", index, "role")
          }
          item = self.create_object(StudyRole, params)
          if item:
            self.items.append(item)
            self.globals.cross_references.add(item.name, item)     
    except Exception as e:
      self._sheet_exception(e)

  def _get_masking(self, index):
    masking = None
    masking_text = self.read_cell_by_name(index, 'masking', default='')
    if masking_text:
      masking = self.create_object(Masking, {'description': masking_text})
    return masking
  
  def _get_refs_for(self, klass, index: int, column_name: str, ids=False):
    collection = []
    refs = self.read_cell_multiple_by_name(index, column_name, must_be_present=False)
    for ref in refs:
      item = self.globals.cross_references.get(klass, ref)
      if item:
        if ids:
          collection.append(item.id)
        else:
          collection.append(item)
      else:
        self._warning(index, column_name, f"Failed to find {klass.__name__.lower()} with name '{ref}'")
    return collection


from typing import List, Literal
from .api_base_model import ApiBaseModelWithIdNameLabelAndDesc
from .study_definition_document_version import StudyDefinitionDocumentVersion
from .code import Code
from .comment_annotation import CommentAnnotation

class StudyDefinitionDocument(ApiBaseModelWithIdNameLabelAndDesc):
  language: Code
  type: Code
  templateName: str
  versions: List[StudyDefinitionDocumentVersion] = []
  notes: List[CommentAnnotation] = []
  instanceType: Literal['StudyDefinitionDocument']



import pandas as pd
from usdm_excel.study_identifiers_sheet.study_identifiers_sheet import StudyIdentifiersSheet
from usdm_model.code import Code

def test_create(mocker, globals):
  globals.cross_references.clear()
  mock_id = mocker.patch("usdm_excel.id_manager.IdManager.build_id")
  mock_id.side_effect=['Code_1', 'Org_1', 'Addr_1', 'Id_1', 'Code_2', 'Org_2', 'Addr_2', 'Id_2', 'Code_3', 'Org_3', 'Addr_3', 'Id_3']
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = [
    [ 'USGOV', 'CT-GOV', 'ClinicalTrials.gov', 'Study Registry', 'NCT12345678', 'line|district|city|state|postal_code|GBR' ],
    [ 'USGOV2', 'CT-GOV2', 'ClinicalTrials2.gov', 'Study Registry', 'NCT12345679', 'line2,district2,city2,state2,postal_code2,FRA' ],
    [ 'USGOV3', 'CT-GOV3', 'ClinicalTrials3.gov', 'Study Registry', 'NCT123456710', 'line3,district3,city3,state3,postal_code3,FR' ]
  ]
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['organisationIdentifierScheme', 'organisationIdentifier', 'organisationName', 'organisationType', 'studyIdentifier', 'organisationAddress'])
  mock_json = mocker.patch("json.load")
  mock_json.return_value = {}
  expected_1 = Code(id='Code1', code='code', codeSystem='codesys', codeSystemVersion='3', decode="GBR")
  expected_2 = Code(id='Code2', code='code', codeSystem='codesys', codeSystemVersion='3', decode="FRA")
  expected_3 = Code(id='Code3', code='code', codeSystem='codesys', codeSystemVersion='3', decode="FRA")
  mock_code = mocker.patch("usdm_excel.iso_3166.ISO3166.code")
  mock_code.side_effect=[expected_1, expected_2, expected_3]
  ids = StudyIdentifiersSheet("", globals)
  assert len(ids.identifiers) == 3
  assert len(ids.organizations) == 3
  assert ids.identifiers[0].id == 'Id_1'
  assert ids.identifiers[0].text == 'NCT12345678'
  assert ids.organizations[0].name == 'ClinicalTrials.gov'
  assert ids.organizations[0].legalAddress.city == 'city'
  assert ids.identifiers[1].id == 'Id_2'
  assert ids.identifiers[1].text == 'NCT12345679'
  assert ids.organizations[1].name == 'ClinicalTrials2.gov'
  assert ids.organizations[1].legalAddress.city == 'city2'
  assert ids.identifiers[2].id == 'Id_3'
  assert ids.identifiers[2].text == 'NCT123456710'
  
def test_create_with_z(mocker, globals):
  globals.cross_references.clear()
  mock_id = mocker.patch("usdm_excel.id_manager.IdManager.build_id")
  mock_id.side_effect=['Code_1', 'Org_1', 'Addr_1', 'Id_1', 'Code_2', 'Org_2', 'Addr_2', 'Id_2', 'Code_3', 'Org_3', 'Addr_3', 'Id_3']
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = [
    [ 'USGOV', 'CT-GOV', 'ClinicalTrials.gov', 'Study Registry', 'NCT12345678', 'line|district|city|state|postal_code|GBR' ],
    [ 'USGOV2', 'CT-GOV2', 'ClinicalTrials2.gov', 'Study Registry', 'NCT12345679', 'line2,district2,city2,state2,postal_code2,FRA' ],
    [ 'USGOV3', 'CT-GOV3', 'ClinicalTrials3.gov', 'Study Registry', 'NCT123456710', 'line3,district3,city3,state3,postal_code3,FR' ]
  ]
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['organizationIdentifierScheme', 'organizationIdentifier', 'organizationName', 'organizationType', 'studyIdentifier', 'organizationAddress'])
  mock_json = mocker.patch("json.load")
  mock_json.return_value = {}
  expected_1 = Code(id='Code1', code='code', codeSystem='codesys', codeSystemVersion='3', decode="GBR")
  expected_2 = Code(id='Code2', code='code', codeSystem='codesys', codeSystemVersion='3', decode="FRA")
  expected_3 = Code(id='Code3', code='code', codeSystem='codesys', codeSystemVersion='3', decode="FRA")
  mock_code = mocker.patch("usdm_excel.iso_3166.ISO3166.code")
  mock_code.side_effect=[expected_1, expected_2, expected_3]
  ids = StudyIdentifiersSheet("", globals)
  assert len(ids.identifiers) == 3
  assert len(ids.organizations) == 3
  assert ids.identifiers[0].id == 'Id_1'
  assert ids.identifiers[0].text == 'NCT12345678'
  assert ids.organizations[0].name == 'ClinicalTrials.gov'
  assert ids.organizations[0].legalAddress.city == 'city'
  assert ids.identifiers[1].id == 'Id_2'
  assert ids.identifiers[1].text == 'NCT12345679'
  assert ids.organizations[1].name == 'ClinicalTrials2.gov'
  assert ids.organizations[1].legalAddress.city == 'city2'
  assert ids.identifiers[2].id == 'Id_3'
  assert ids.identifiers[2].text == 'NCT123456710'
  
def test_create_new_columns(mocker, globals):
  globals.cross_references.clear()
  mock_id = mocker.patch("usdm_excel.id_manager.IdManager.build_id")
  mock_id.side_effect=['Code_1', 'Org_1', 'Addr_1', 'Id_1', 'Code_2', 'Org_2', 'Addr_2', 'Id_2', 'Code_3', 'Org_3', 'Addr_3', 'Id_3']
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = [
    [ 'USGOV', 'CT-GOV', 'ClinicalTrials.gov', 'CT.gov [1]', 'Study Registry', 'NCT12345678', 'line|district|city|state|postal_code|GBR' ],
    [ 'USGOV2', 'CT-GOV2', 'ClinicalTrials2.gov', '', 'Study Registry', 'NCT12345679', 'line2,district2,city2,state2,postal_code2,FRA' ],
    [ 'USGOV3', 'CT-GOV3', 'ClinicalTrials3.gov', 'CT.gov [3]', 'Study Registry', 'NCT123456710', 'line3,district3,city3,state3,postal_code3,FR' ]
  ]
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['IdentifierScheme', 'organisationIdentifier', 'name', 'label', 'type', 'studyIdentifier', 'address'])
  mock_json = mocker.patch("json.load")
  mock_json.return_value = {}
  expected_1 = Code(id='Code1', code='code', codeSystem='codesys', codeSystemVersion='3', decode="GBR")
  expected_2 = Code(id='Code2', code='code', codeSystem='codesys', codeSystemVersion='3', decode="FRA")
  expected_3 = Code(id='Code3', code='code', codeSystem='codesys', codeSystemVersion='3', decode="FRA")
  mock_code = mocker.patch("usdm_excel.iso_3166.ISO3166.code")
  mock_code.side_effect=[expected_1, expected_2, expected_3]
  ids = StudyIdentifiersSheet("", globals)
  assert len(ids.identifiers) == 3
  assert ids.identifiers[0].id == 'Id_1'
  assert ids.identifiers[0].text == 'NCT12345678'
  assert ids.organizations[0].name == 'ClinicalTrials.gov'
  assert ids.organizations[0].label == 'CT.gov [1]'
  assert ids.organizations[0].legalAddress.city == 'city'
  assert ids.identifiers[1].id == 'Id_2'
  assert ids.identifiers[1].text == 'NCT12345679'
  assert ids.organizations[1].name == 'ClinicalTrials2.gov'
  assert ids.organizations[1].legalAddress.city == 'city2'
  assert ids.organizations[1].label == ''
  assert ids.identifiers[2].id == 'Id_3'
  assert ids.identifiers[2].text == 'NCT123456710'
  
def test_create_empty(mocker, globals):
  globals.cross_references.clear()
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = []
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['studyIdentifierName', 'studyIdentifierDescription', 'studyIdentifierType'])
  ids = StudyIdentifiersSheet("", globals)
  assert len(ids.identifiers) == 0

def test_read_cell_by_name_error(mocker, globals):
  globals.cross_references.clear()
  mock_error = mocker.patch("usdm_excel.errors_and_logging.errors.Errors.add")
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = [['Id 1', 'Id One']]
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['studyIdentifierName', 'studyIdentifierDescription'])
  ids = StudyIdentifiersSheet("", globals)
  mock_error.assert_called()
  assert mock_error.call_args[0][0] == "studyIdentifiers"
  assert mock_error.call_args[0][1] == None
  assert mock_error.call_args[0][2] == None
  assert mock_error.call_args[0][3] == "Exception. Error [Failed to detect column(s) 'organisationType, organizationType, type' in sheet] while reading sheet 'studyIdentifiers'. See log for additional details."
  
def test_address_error(mocker, globals):
  globals.cross_references.clear()
  mock_error = mocker.patch("usdm_excel.errors_and_logging.errors.Errors.add")
  mock_id = mocker.patch("usdm_excel.id_manager.IdManager.build_id")
  mock_id.side_effect=['Code_1', 'Org_1', 'Addr_1', 'Id_1']
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = [
    [ 'USGOV', 'CT-GOV', 'ClinicalTrials.gov', 'Study Registry', 'NCT12345678', 'line|city|district|state|GBR' ],
  ]
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['organisationIdentifierScheme', 'organisationIdentifier', 'organisationName', 'organisationType', 'studyIdentifier', 'organisationAddress'])
  mock_json = mocker.patch("json.load")
  mock_json.return_value = {}
  expected_1 = Code(id='Code1', code='code', codeSystem='codesys', codeSystemVersion='3', decode="GBR")
  mock_code = mocker.patch("usdm_excel.iso_3166.ISO3166.code")
  mock_code.side_effect=[expected_1]
  ids = StudyIdentifiersSheet("", globals)
  mock_error.assert_called()
  assert mock_error.call_args[0][0] == "studyIdentifiers"
  assert mock_error.call_args[0][1] == 1
  assert mock_error.call_args[0][2] == 6
  assert mock_error.call_args[0][3] == "Address 'line|city|district|state|GBR' does not contain the required fields (lines, district, city, state, postal code and country code) using '|' separator characters, only 5 found"



from usdm_excel.ct_version_manager import *

def test_create(globals):
  object = CTVersionManager(globals)
  assert len(object._items.keys()) == 0
  assert object._items == {}

def test_add(globals):
  globals.ct_version_manager._items = {}
  globals.ct_version_manager.add('fred', 'value')
  assert len(globals.ct_version_manager._items.keys()) == 1
  assert globals.ct_version_manager._items['FRED'] == 'value'
  globals.ct_version_manager.add('Sid1', 'value')
  assert len(globals.ct_version_manager._items.keys()) == 2
  assert globals.ct_version_manager._items['SID1'] == 'value'

def test_get(globals):
  globals.ct_version_manager._items = {}
  globals.ct_version_manager._items['FRED'] = 'value'
  assert globals.ct_version_manager.get('fred') == 'value'
  assert globals.ct_version_manager.get('FRED') == 'value'

def test_clear(globals):
  globals.ct_version_manager._items = {}
  globals.ct_version_manager._items['fred'] = 'value'
  assert len(globals.ct_version_manager._items.keys()) == 1
  globals.ct_version_manager.clear()
  assert len(globals.ct_version_manager._items.keys()) == 0



from tests.mocks.mock_general import *
from tests.mocks.mock_sheet import *
from tests.mocks.mock_ids import *
from tests.mocks.mock_logging import *
from usdm_model.study_change import StudyChange
from usdm_model.document_content_reference import DocumentContentReference
from usdm_excel.study_amendment_sheet.study_amendment_changes_sheet import StudyAmendmentChangesSheet
from usdm_model.code import Code

def test_create(mocker, globals):
  sheet_data = {
    'amendment': ['A1', 'A2', 'A3'], 
    'name': ['Change One', 'Change Two', 'Change Three'], 
    'description': ['Change Desc One', 'Change Desc Two', 'Change Desc Three'], 
    'label': ['Change Label One', 'Change Label Two', 'Change Label Three'],
    'rationale': ['R1', 'R2', 'R3'],
    'summary': ['Summary 1', 'Summary 2', 'Summary 3'],
    'sections': ['1.2.1: XXX', '1.3.1: YYY, 3.2: ZZZ', '4.5: Something Long']
  }
  index = 1
  object_data = []
  for y in range(3):
    dcrs = []
    refs = sheet_data['sections'][y]
    for ref in refs.split(','):
      parts = ref.split(':')
      dcr = {'sectionNumber': parts[0].strip(), 'sectionTitle': parts[1].strip()}
      dcrs.append(fake_create_object(DocumentContentReference, dcr, index))
      sc = {'cls': DocumentContentReference, 'data': dcr}
      object_data.append(sc)    
      index += 1
    sc = {'cls': StudyChange, 'data': {'changedSections': dcrs}}
    for x in ['name', 'description', 'label', 'rationale', 'summary']:
      sc['data'][x] = sheet_data[x][y]
    object_data.append(sc)
  mock_create_object(mocker, object_data)
  mock_sheet_present(mocker)
  mock_sheet(mocker, globals, sheet_data)
  item = StudyAmendmentChangesSheet("", globals)
  assert len(item.items) == 3
  assert item.items[0].model_dump() == {
    'id': 'StudyChange_2',
    'description': 'Change Desc One',
    'instanceType': 'StudyChange',
    'label': 'Change Label One',
    'name': 'Change One',
    'rationale': 'R1',
    'summary': 'Summary 1',
    'changedSections': [
      {
        'id': 'DocumentContentReference_1',
        'instanceType': 'DocumentContentReference',
        'sectionNumber': '1.2.1',
        'sectionTitle': 'XXX',
      },
    ]
  }
  assert item.items[1].model_dump() == {
    'id': 'StudyChange_5',
    'description': 'Change Desc Two',
    'instanceType': 'StudyChange',
    'label': 'Change Label Two',
    'name': 'Change Two',
    'rationale': 'R2',
    'summary': 'Summary 2',
    'changedSections': [
      {
        'id': 'DocumentContentReference_2',
        'instanceType': 'DocumentContentReference',
        'sectionNumber': '1.3.1',
        'sectionTitle': 'YYY',
      },
      {
        'id': 'DocumentContentReference_3',
        'instanceType': 'DocumentContentReference',
        'sectionNumber': '3.2',
        'sectionTitle': 'ZZZ',
      },
    ]
  }
  assert item.items[2].model_dump() == {
    'id': 'StudyChange_7',
    'description': 'Change Desc Three',
    'instanceType': 'StudyChange',
    'label': 'Change Label Three',
    'name': 'Change Three',
    'rationale': 'R3',
    'summary': 'Summary 3',
    'changedSections': [
      {
        'id': 'DocumentContentReference_4',
        'instanceType': 'DocumentContentReference',
        'sectionNumber': '4.5',
        'sectionTitle': 'Something Long',
      },
    ]
  }
  
def test_create_empty(mocker, globals):
  sheet_data = {}
  mock_sheet_present(mocker)
  mock_sheet(mocker, globals, sheet_data)
  item = StudyAmendmentChangesSheet("", globals)
  assert len(item.items) == 0

def test_read_cell_by_name_error(mocker, globals):
  sheet_data = {
    'amendment': ['A1'], 
    'name': ['Change One'], 
    'description': ['Change Desc One'], 
    'label': ['Change Label One'],
    'summary': ['Summary 1'],
    'sections': ['1.2.1: XXX']
  }
  mea = mock_error_add(mocker, [None, None, None, None, None, None])
  mock_sheet_present(mocker)
  mock_sheet(mocker, globals, sheet_data)
  item = StudyAmendmentChangesSheet("", globals)
  assert mock_called(mea, 2)
  mock_parameters_correct(mea, [mocker.call('amendmentChanges', 1, -1, "Error attempting to read cell 'rationale'. Exception: Failed to detect column(s) 'rationale' in sheet", 40), 
                                mocker.call('amendmentChanges', 1, 1, "Failed to find amendment 'A1'", 40)])



import os
import yaml
import requests
from usdm_excel.errors_and_logging.errors_and_logging import ErrorsAndLogging
from usdm_excel.cdisc_ct_library import CDISCCTLibrary
from usdm_excel.id_manager import IdManager
from usdm_model.biomedical_concept import BiomedicalConcept
from usdm_model.biomedical_concept_property import BiomedicalConceptProperty
from usdm_model.response_code import ResponseCode
from usdm_model.code import Code
from usdm_model.alias_code import AliasCode

class CDISCBCLibrary():

  API_ROOT = 'https://api.library.cdisc.org/api/cosmos/v2'    
  
  def __init__(self, errors_and_logging: ErrorsAndLogging, ct_library: CDISCCTLibrary, id_manager: IdManager):
    self._errors_and_logging = errors_and_logging
    self._ct_library = ct_library
    self._id_manager = id_manager
    self._api_key = os.getenv('CDISC_API_KEY')
    self._headers =  {"Content-Type":"application/json", "api-key": self._api_key}
    self._package_metadata = {}
    self._package_items = {}
    self._bc_responses = {}
    self._bcs = {}
    self._bcs_raw = {}
    self._bc_index = {}
    self._map = {}
    if self._bcs_exist():
      self._bcs = self._load_bcs()
    else:
      self._get_package_metadata()
      self._get_package_items()
      self._get_sdtm_bcs()
      self._get_generic_bcs()
      self._save_bcs(self._bcs_raw)
    self._bc_index = self._create_bc_index()

  def exists(self, name):
    return True if name.upper() in self._bc_index else False

  def catalogue(self) -> list:
    return list(self._bcs.keys())
  
  def usdm(self, name) -> BiomedicalConcept:
    bc = self._get_bc_data(name) if self.exists(name) else None
    if bc:
      bc_copy = bc.model_dump()
      self._set_ids(bc_copy)
      bc = BiomedicalConcept(**bc_copy)
    return bc

  def _set_ids(self, parent):
    if isinstance(parent, str) or isinstance(parent, bool):
      return
    parent['id'] = self._id_manager.build_id(parent['instanceType'])
    for name, value in parent.items():
      if isinstance(value, list):
        for child in value:
          self._set_ids(child)    
      else:
        self._set_ids(value)

  def _load_bcs(self):
    results = {}
    data = self._read_bcs()
    for key, item in data.items():
      results[key] = BiomedicalConcept(**item)
    return results
  
  def _create_bc_index(self):
    results = {}
    for name, item in self._bcs.items():
      results[name] = name
      for synonym in item.synonyms:
        results[synonym.upper()] = name
    return results

  def _get_bc_data(self, name):
    return self._bcs[self._bc_index[name.upper()]]
  
  def _get_package_metadata(self):
    urls = {
      'generic': '/mdr/bc/packages',
      'sdtm': '/mdr/specializations/sdtm/packages'
    } 
    for url_type, url in urls.items():
      try:
        api_url = self._url(url)
        self._logger.info(f"CDISC BC Library: {url_type}: {url}")
        raw = requests.get(api_url, headers=self._headers)
        response = raw.json()
        self._package_metadata[url_type] = response['_links']['packages']
      except Exception as e:
        self._exception(f"Exception '{e}', failed to retrieve CDISC BC package metadata from '{api_url}'", e)
    self._logger.debug(f"PACKAGES: {self._package_metadata}")
        
  def _get_package_items(self) -> dict:
    #self._package_items, self._map
    for package_type in ['sdtm', 'generic']:
      self._package_items[package_type] = {}
      for package in self._package_metadata[package_type]:
        self._get_package(package, package_type)        

  def _get_package(self, package, package_type):
    try:
      response_field = {'sdtm': 'datasetSpecializations', 'generic': 'biomedicalConcepts'}
      api_url = self._url(package['href']) if 'href' in package else 'not set'
      self._logger.info(f"CDISC BC Library: {package_type}, {api_url}")
      raw = requests.get(api_url, headers=self._headers)
      response = raw.json()
      for item in response['_links'][response_field[package_type]]:
        self._logger.debug(f"ITEM: {item}")
        key = item['title'].upper()
        if package_type == 'sdtm':
          self._package_items[package_type][key] = item
          #map[item['href']] = key
        elif package_type == 'generic' and not key in self._package_items:
          self._logger.info(f"GENERIC: Detected generic only BC {key}")
          self._package_items[package_type][key] = item
          self._map[item['href']] = key
    except Exception as e:
      self._exception(f"Exception '{e}', failed to retrieve CDISC BC metadata from '{api_url}'", e)
      return {}

  def _get_sdtm_bcs(self):
    # self._bcs, self._bcs_raw = 
    for name, item in self._package_items['sdtm'].items():
      #print(f"ITEM IN GET BC: {item}")
      sdtm, generic = self._get_from_url_all(name, item)
      if sdtm:
        bc = self._sdtm_bc_as_usdm(sdtm, generic)
        if bc:
          if 'variables' in sdtm:
            for item in sdtm['variables']:
                property = self._sdtm_bc_property_as_usdm(item, generic)
                if property:
                  bc.properties.append(property)
          self._bcs_raw[name] = bc.model_dump()
          self._bcs[name] = bc
        if generic:
          href = generic['_links']['self']['href']
          if href in self._map:
            #print(f"POP: {href}")
            self._map.pop(generic['_links']['self']['href'])
          else:
            self._logger.error(f"Missing reference when popping {href}")
    #print(f"REMAINING: {self._map}")

  def _get_generic_bcs(self) -> BiomedicalConcept:
    #print(f"GENRIC KEYS: {self._package_items['generic'].keys()}")
    for name, item in self._package_items['generic'].items():
      if self._process_genric_bc(name):
        #print(f"ITEM IN GET GENERIC BC: {item}")
        response = self._get_from_url(item['href'])
        bc = self._generic_bc_as_usdm(response)
        if 'dataElementConcepts' in response:
          for item in response['dataElementConcepts']:
            property = self._generic_bc_property_as_usdm(item)
            if property:
              bc.properties.append(property)
        self._bcs_raw[name] = bc.model_dump()
        self._bcs[name] = bc

  def _generic_bc_as_usdm(self, api_bc) -> BiomedicalConcept:
    concept_code = self._cdisc_code(api_bc['conceptId'], api_bc['shortName'])
    synonyms = api_bc['synonyms'] if 'synonyms' in api_bc else []
    return self._biomedical_concept_object(api_bc['shortName'], api_bc['shortName'], synonyms, api_bc['_links']['self']['href'], concept_code)

  def _generic_bc_property_as_usdm(self, property) -> BiomedicalConceptProperty:
    concept_code = self._cdisc_code(property['conceptId'], property['shortName'])
    responses = []
    if 'exampleSet' in property:
      for example in property['exampleSet']:
        term = self._ct_library.preferred_term(example)
        if term != None:
          code = self._cdisc_code(term['conceptId'], term['preferredTerm'])
          code.id = "tbd"
          responses.append(ResponseCode(id="tbd", isEnabled=True, code=code))
    return self._biomedical_concept_property_object(property['shortName'], property['shortName'], property['dataType'], responses, concept_code)

  def _sdtm_bc_as_usdm(self, sdtm, generic) -> BiomedicalConcept:
    try:
      if self._process_sdtm_bc(sdtm['shortName']):
        self._logger.debug(f"BC: {sdtm}\n\n{generic}")
        role_variable = self._get_role_variable(sdtm)
        if role_variable:
          if 'assignedTerm' in role_variable:
            if 'conceptId' in role_variable['assignedTerm'] and 'value' in role_variable['assignedTerm']:
              concept_code = self._cdisc_code(role_variable['assignedTerm']['conceptId'], role_variable['assignedTerm']['value'])
            else:
              self._logger.error(f"Failed to set BC concept 1, {sdtm['shortName']}")
              concept_code = self._cdisc_code('No Concept Code', role_variable['assignedTerm']['value'])
          else:
            self._logger.error(f"Failed to set BC concept 2, {sdtm['shortName']}")
            concept_code = self._cdisc_code(generic['conceptId'], generic['shortName'])
        else:
          self._logger.error(f"Failed to set BC concept {sdtm['shortName']}")
          concept_code = self._cdisc_code(generic['conceptId'], generic['shortName'])
        synonyms = generic['synonyms'] if 'synonyms' in generic else []
        synonyms.append(generic['shortName'])
        return self._biomedical_concept_object(sdtm['shortName'], sdtm['shortName'], synonyms, sdtm['_links']['self']['href'], concept_code)
      else:
        return None
    except Exception as e:
      self._exception(f"Exception '{e}', failed to build BC {sdtm['shortName']}", e)
      return None

  def _sdtm_bc_property_as_usdm(self, sdtm_property, generic) -> BiomedicalConceptProperty:
    try:
      self._logger.debug(f"NAME: {sdtm_property['name']}, {sdtm_property['name'][2:]}")
      if self._process_property(sdtm_property['name']):
        self._logger.debug(f"PROPERTY: {sdtm_property}")
        if 'dataElementConceptId' in sdtm_property:
          generic_match = self._get_dec_match(generic, sdtm_property['dataElementConceptId'])
          if generic_match:
            concept_code = self._cdisc_code(generic_match['conceptId'], generic_match['shortName'])
          else:
            if 'assignedTerm' in sdtm_property and 'conceptId' in sdtm_property['assignedTerm'] and 'value' in sdtm_property['assignedTerm']:
              concept_code = self._cdisc_code(sdtm_property['dataElementConceptId'], sdtm_property['name'])
            else:
              self._logger.error(f"Failed to set property concept 1, {sdtm_property}")
              concept_code = self._cdisc_code(sdtm_property['dataElementConceptId'], sdtm_property['name'])
        else:
          if 'assignedTerm' in sdtm_property:
            concept_code = self._cdisc_code(sdtm_property['assignedTerm']['conceptId'], sdtm_property['assignedTerm']['value'])
          else:
            self._logger.error(f"Failed to set property concept 2, {sdtm_property}")
            concept_code = self._cdisc_code('No Concept Code', sdtm_property['name'])
        responses = []
        codes = []
        if 'valueList' in sdtm_property:
          codelist = sdtm_property['codelist']['conceptId'] if 'codelist' in sdtm_property else None
          for value in sdtm_property['valueList']:
            term = self._ct_library.preferred_term(value, codelist)
            if term:
              code = self._cdisc_code(term['conceptId'], term['preferredTerm'])
              code.id = "tbd"
              codes.append(code)
            else:
              term = self._ct_library.submission(value, codelist)
              if term:
                code = self._cdisc_code(term['conceptId'], term['preferredTerm'])
                code.id = "tbd"
                codes.append(code)
              else:
                cl = f", code list {sdtm_property['codelist']['conceptId'] if 'codelist' in sdtm_property else '<not defined>'}"
                self._logger.error(f"Failed to find submission or preferred term '{value}' {cl}")
        for code in codes:
         response_code = ResponseCode(id="tbd", isEnabled=True, code=code)
         responses.append(response_code)
        datatype = sdtm_property['dataType'] if 'dataType' in sdtm_property else ''
        return self._biomedical_concept_property_object(sdtm_property['name'], sdtm_property['name'], datatype, responses, concept_code)
      else:
        return None
    except Exception as e:
      self._exception(f"Exception '{e}', failed to build property {sdtm_property}", e)
      return None

  def _biomedical_concept_object(self, name, label, synonyms, reference, code) -> BiomedicalConcept:
    code.id = "tbd"
    alias_code=self._alias_code(code, [])
    alias_code.id = "tbd"
    return BiomedicalConcept(
      id="tbd",
      name=name,
      label=label,
      synonyms=synonyms,
      reference=reference,
      properties=[],
      code=alias_code
    )

  def _biomedical_concept_property_object(self, name, label, datatype, responses, code):
    code.id = "tbd"
    alias_code=self._alias_code(code, [])
    alias_code.id = "tbd"
    return BiomedicalConceptProperty(
      id="tbd",
      name=name,
      label=label,
      isRequired=True,
      isEnabled=True,
      datatype=datatype,
      responseCodes=responses,
      code=alias_code
    )

  def _process_sdtm_bc(self, name):
    if name in [
      'Exclusion Criteria 01', 'Inclusion Criteria 01', "Medical History Prespecified: Alzheimer's Disease", "Medical History Prespecified: Confusional Episodes", 
      "Medical History Prespecified: Essential Tremor",
      "Medical History Prespecified: Extrapyramidal Features",
      "Medical History Prespecified: Facial Masking",
      "Medical History Prespecified: Rigidity Upper Extremity",
      "Medical History Prespecified: Sensitivity to Neuroleptics",
      "Medical History Prespecified: Visual Hallucinations",
      "TTS Acceptability Survey - Patch Acceptability",
      "TTS Acceptability Survey - Patch Appearance",
      "TTS Acceptability Survey - Patch Durability",
      "TTS Acceptability Survey - Patch Size",
      "Beer Use History",
      "Cigarette History",
      "Cigar History",
      "Coffee Use History",
      "Cola Use History",
      "Distilled Spirits Use History",
      "Pipe History",
      "Tea Use History",
      "Wine Use History"
    ]:
      return False
    return True
  
  def _process_genric_bc(self, name):
    return True if name.upper() in [ "SUBJECT AGE", "RACE", "SEX" ] else False
  
  def _process_property(self, name):
    if name[2:] in [
      'TEST', 'STRESN', 'STRESU', 'STRESC', 'CLASCD', 'LOINC', 'LOT', 'CAT', 'SCAT', 'LLT', 'LLTCD', 'HLT', 'HLTCD',
      'PTCD', 'BODSYS', 'BDSYCD', 'SOC', 'SOCCD', 'RLDEV'
    ]:
      return False
    if name in ['EPOCH']:
      return False
    return True
 
  def _get_role_variable(self, data):
    return next((item for item in data['variables'] if item["role"] == "Topic"), None)

  def _get_dec_match(self, data, id):
    return next((item for item in data['dataElementConcepts'] if item["conceptId"] == id), None)

  def _get_from_url_all(self, name, details) -> dict:
    try:
      self._logger.debug(f"DETAILS: {details}")
      sdtm_response = self._get_from_url(details['href'])
      generic = sdtm_response["_links"]["parentBiomedicalConcept"]
      generic_response = self._get_from_url(generic['href'])
      return sdtm_response, generic_response
    except Exception as e:
      self._exception(f"Exception '{e}', failed to retrieve CDISC BC metadata for {name} from '{details['href']}'", e)
      return None, None

  def _get_from_url(self, url):
    api_url = self._url(url)
    self._logger.info("CDISC BC Library: %s" % api_url)
    raw = requests.get(api_url, headers=self._headers)
    result = raw.json()
    return result

  def _url(self, relative_url) -> str:
    return "%s%s" % (self.__class__.API_ROOT, relative_url)

  def _save_bcs(self, data):
    try:
      with open(self._bcs_filename(), 'w') as f:
        yaml.dump(data, f, indent=2, sort_keys=True)
    except Exception as e:
      self._exception(f"Exception '{e}', failed to save CDSIC BC file", e)

  def _read_bcs(self):
    try:
      if self._bcs_exist():
        with open(self._bcs_filename()) as f:
          return yaml.load(f, Loader=yaml.FullLoader)
      else:
        self._logger.error(f"Failed to read CDSIC BC file, does not exist")
        return None
    except Exception as e:
      self._exception(f"Exception '{e}', failed to read CDSIC CT file", e)

  def _bcs_exist(self):
    return os.path.isfile(self._bcs_filename()) 

  def _bcs_filename(self):
    return os.path.join(os.path.dirname(__file__), 'data', f"cdisc_bcs.yaml")

  def _cdisc_code(self, code, decode):
    id = self._id_manager.build_id(Code)
    instance = Code(id=id, code=code, codeSystem=self._library.system, codeSystemVersion=self._library.version, decode=decode)
    self._cross_references.add(instance.id, instance)
    return instance

  def _alias_code(self, standard_code, aliases):
    return AliasCode(id=self._id_manager.build_id(AliasCode), standardCode=standard_code, standardCodeAliases=aliases) if standard_code else None  
  
  def _exception(self, message, e):
    self._logger.error(message)
    self._logger.error(f"{e}\n{traceback.format_exc()}")




from typing import List, Literal
from .api_base_model import ApiBaseModelWithId
from .code import Code
from .governance_date import GovernanceDate
from .narrative_content import NarrativeContent
from .comment_annotation import CommentAnnotation

class StudyDefinitionDocumentVersion(ApiBaseModelWithId):
  version: str
  status: Code
  dateValues: List[GovernanceDate] = []
  contents: List[NarrativeContent] = []
  childIds: List[str] = []
  notes: List[CommentAnnotation] = []
  instanceType: Literal['StudyDefinitionDocumentVersion']

  def narrative_content_in_order(self):
    sections = []
    narrative_content = self._first_narrative_content()
    while narrative_content:
      sections.append(narrative_content)
      narrative_content = next((x for x in self.contents if x.id == narrative_content.nextId), None)
    return sections
  
  def _first_narrative_content(self) -> NarrativeContent:
    return next((x for x in self.contents if not x.previousId and x.nextId), None)
  



from typing import List, Literal, Union
from .api_base_model import ApiBaseModelWithIdNameLabelAndDesc
from .comment_annotation import CommentAnnotation

class BiomedicalConceptSurrogate(ApiBaseModelWithIdNameLabelAndDesc):
  reference: Union[str, None] = None
  notes: List[CommentAnnotation] = []
  instanceType: Literal['BiomedicalConceptSurrogate']



from typing import List, Literal, Union
from .api_base_model import ApiBaseModelWithIdNameLabelAndDesc
from .procedure import Procedure
from .comment_annotation import CommentAnnotation

class Activity(ApiBaseModelWithIdNameLabelAndDesc):
  previousId: Union[str, None] = None
  nextId: Union[str, None] = None
  childIds: List[str] = []
  definedProcedures: List[Procedure] = []
  biomedicalConceptIds: List[str] = []
  bcCategoryIds: List[str] = []
  bcSurrogateIds: List[str] = []
  timelineId: Union[str, None] = None
  notes: List[CommentAnnotation] = []
  instanceType: Literal['Activity']



from typing import List, Literal, Union
from pydantic import Field
from .api_base_model import ApiBaseModel
from .study_definition_document import StudyDefinitionDocument
from .study_version import StudyVersion
from uuid import UUID

class Study(ApiBaseModel):
  id: Union[UUID, None] = None
  name: str = Field(min_length=1)
  description: Union[str, None] = None
  label: Union[str, None] = None
  versions: List[StudyVersion] = []
  documentedBy: List[StudyDefinitionDocument] = []
  instanceType: Literal['Study']

  def document_by_template_name(self, name: str) -> StudyDefinitionDocument:
    # for x in self.documentedBy:
    #   print(f"DOC: {x.templateName}, {name}")
    return next((x for x in self.documentedBy if x.templateName.upper() == name.upper()), None)


from typing import Literal, Union
from .api_base_model import ApiBaseModelWithId
from .code import Code

class Range(ApiBaseModelWithId):
  minValue: float
  maxValue: float
  unit: Union[Code, None] = None
  isApproximate: bool
  instanceType: Literal['Range']



import traceback
from usdm_excel.base_sheet import BaseSheet
from usdm_model.study_epoch import StudyEpoch
from usdm_excel.globals import Globals

class StudyDesignEpochSheet(BaseSheet):

  SHEET_NAME = 'studyDesignEpochs'
  
  def __init__(self, file_path: str, globals: Globals):
    try:
      self.items = []
      super().__init__(file_path=file_path, globals=globals, sheet_name=self.SHEET_NAME)
      for index, row in self.sheet.iterrows():
        name = self.read_cell_by_name(index, ['studyEpochName', 'name'])
        description = self.read_cell_by_name(index, ['studyEpochDescription', 'description'])
        label = self.read_cell_by_name(index, 'label', default="", must_be_present=False)
        epoch_type = self.read_cdisc_klass_attribute_cell_by_name('StudyEpoch', 'studyEpochType', index, ['studyEpochType', 'type'])
        notes = self.read_cell_multiple_by_name(index, 'notes', must_be_present=False)
        item = self.create_object(StudyEpoch, {'name': name, 'description': description, 'label': label, 'type': epoch_type})
        if item:
          self.items.append(item)
          self.globals.cross_references.add(name, item)
          self.add_notes(item, notes)
    except Exception as e:
      self._sheet_exception(e)


from typing import Union
from .api_base_model import ApiBaseModel
from .study import Study

class Wrapper(ApiBaseModel):
  study: Study
  usdmVersion: str
  systemName: Union[str, None] = None
  systemVersion: Union[str, None] = None


from typing import List, Literal
from .api_base_model import ApiBaseModelWithIdNameLabelAndDesc
from .comment_annotation import CommentAnnotation

class AnalysisPopulation(ApiBaseModelWithIdNameLabelAndDesc):
  text: str
  notes: List[CommentAnnotation] = []
  instanceType: Literal['AnalysisPopulation']



from usdm_model.study import Study
from usdm_model.narrative_content import NarrativeContent, NarrativeContentItem
from usdm_db.cross_reference import CrossReference
from usdm_db.errors_and_logging.errors_and_logging import ErrorsAndLogging
from usdm_db.document.utility import get_soup
from fhir.resources.bundle import Bundle, BundleEntry
from fhir.resources.identifier import Identifier
from fhir.resources.composition import Composition, CompositionSection
from fhir.resources.narrative import Narrative
from fhir.resources.codeableconcept import CodeableConcept
from fhir.resources.reference import Reference
from uuid import uuid4
import datetime

class ToFHIR():

  EMPTY_DIV = '<div xmlns="http://www.w3.org/1999/xhtml"></div>'

  class LogicError(Exception):
    pass

  def __init__(self, study: Study, template: str):
    self._study = study
    self._errors_and_logging = ErrorsAndLogging()
    self._cross_ref = CrossReference(study, self._errors_and_logging)
    self._study_version = self._study.versions[0]
    self._study_design = self._study_version.studyDesigns[0]
    self._document = self._study.document_by_template_name(template)
    self._document_version = self._document.versions[0]
    self._doc_title = self._study_version.get_title('Official Study Title')
    self._doc_title = self._doc_title.text if self._doc_title else '[Missing Doc Title]'
    self._map = {x.id:x for x in self._study_version.narrativeContentItems}

  def to_fhir(self, uuid: uuid4):
    try:
      sections = []
      nc_list = self._document_version.narrative_content_in_order()
      for narrative_content in nc_list:
        narrative_content_item = self._map[narrative_content.contentItemId] if narrative_content.contentItemId else None
        sections.append(self._content_to_section(narrative_content, narrative_content_item.text if narrative_content_item else self.EMPTY_DIV))
        narrative_content = next((x for x in self._document_version.contents if x.id == narrative_content.nextId), None)
      type_code = CodeableConcept(text=f"EvidenceReport")
      #date = datetime.datetime.now().isoformat()
      date = datetime.datetime.now(tz=datetime.timezone.utc).isoformat()
      author = Reference(display="USDM")
      print(f"TITLE: {self._doc_title}")
      composition = Composition(title=self._doc_title, type=type_code, section=sections, date=date, status="preliminary", author=[author])
      identifier = Identifier(system='urn:ietf:rfc:3986', value=f'urn:uuid:{uuid}')
      bundle_entry = BundleEntry(resource=composition, fullUrl="https://www.example.com/Composition/1234")
      bundle = Bundle(id=None, entry=[bundle_entry], type="document", identifier=identifier, timestamp=date)
      return bundle.json()
    except Exception as e:
      self._errors_and_logging.exception(f"Exception raised generating FHIR content. See logs for more details", e)
      return None

  def _content_to_section(self, content: NarrativeContent, item_text: str) -> CompositionSection:
    div = self._translate_references(item_text)
    #text = self._add_section_heading(content, div)
    text = str(div)
    text = self._remove_line_feeds(text)
    narrative = Narrative(status='generated', div=text)
    title = self._format_section_title(content.sectionTitle)
    code = CodeableConcept(text=f"section{content.sectionNumber}-{title}")
    section = CompositionSection(title=content.sectionTitle, code=code, text=narrative, section=[])
    return section

  def _format_section_title(self, title: str) -> str:
    return title.lower().strip().replace(' ', '-')
  
  def _clean_section_number(self, section_number: str) -> str:
    return section_number[:-1] if section_number.endswith('.') else section_number
  
  def _translate_references(self, content_text: str):
    soup = get_soup(content_text, self._errors_and_logging)
    for ref in soup(['usdm:ref']):
      try:
        attributes = ref.attrs
        instance = self._cross_ref.get(attributes['klass'], attributes['id'])
        value = self._resolve_instance(instance, attributes['attribute'])
        translated_text = self._translate_references(value)
        self._replace_and_highlight(ref, translated_text)
      except Exception as e:
        self._errors_and_logging.exception(f"Exception raised while attempting to translate reference '{attributes}' while generating the FHIR message, see the logs for more info", e)
        self._replace_and_highlight(ref, 'Missing content: exception')
    self._errors_and_logging.debug(f"Translate references from {content_text} => {get_soup(str(soup), self._errors_and_logging)}")
    return get_soup(str(soup), self._errors_and_logging)

  def _resolve_instance(self, instance, attribute):
    dictionary = self._get_dictionary(instance)
    value = str(getattr(instance, attribute))
    soup = get_soup(value, self._errors_and_logging)
    for ref in soup(['usdm:tag']):
      try:
        attributes = ref.attrs
        if dictionary:
          entry = next((item for item in dictionary.parameterMaps if item.tag == attributes['name']), None)
          if entry:
            self._replace_and_highlight(ref, get_soup(entry.reference, self._errors_and_logging))
          else:
            self._errors_and_logging.error(f"Missing dictionary entry while attempting to resolve reference '{attributes}' while generating the FHIR message")
            self._replace_and_highlight(ref, 'Missing content: missing dictionary entry')
        else:
          self._errors_and_logging.error(f"Missing dictionary while attempting to resolve reference '{attributes}' while generating the FHIR message")
          self._replace_and_highlight(ref, 'Missing content: missing dictionary')
      except Exception as e:
        self._errors_and_logging.exception(f"Failed to resolve reference '{attributes} while generating the FHIR message", e)
        self._replace_and_highlight(ref, 'Missing content: exception')
    return str(soup)

  def _replace_and_highlight(self, ref, text):
    ref.replace_with(text)

  def _get_dictionary(self, instance):
    try:
      return self._cross_ref.get('SyntaxTemplateDictionary', instance.dictionaryId)
    except:
      return None  

  # def _add_section_heading(self, content: NarrativeContent, div) -> str:
  #   DIV_OPEN_NS = '<div xmlns="http://www.w3.org/1999/xhtml">'
  #   text = str(div)
  #   text = text.replace(DIV_OPEN_NS, f"{DIV_OPEN_NS}<p>{content.sectionNumber} {content.sectionTitle}</p>")
  #   return text

  def _remove_line_feeds(self, div: str) -> str:
    #print(f"LB: {len(div)}")
    text = div.replace('\n', '')
    #print(f"LA: {len(text)}")
    return text



from usdm_excel.base_sheet import BaseSheet
from usdm_excel.study_soa_v2_sheet.soa_column_rows import SoAColumnRows
from usdm_excel.study_soa_v2_sheet.conditons import Conditons
#from usdm_excel.id_manager import id_manager
#from usdm_excel.cross_ref import cross_references
from usdm_model.scheduled_instance import ScheduledActivityInstance, ScheduledDecisionInstance
from usdm_model.encounter import Encounter
from usdm_model.study_epoch import StudyEpoch
from usdm_model.activity import Activity
import traceback

class ScheduledInstance():
  
  def __init__(self, parent, col_index):
    self.parent = parent
    self.item = None
    self.col_index = col_index
    epoch_id = None
    encounter_id = None
    name = self.parent.read_cell(SoAColumnRows.NAME_ROW, col_index)
    self.name = name
    description = self.parent.read_cell(SoAColumnRows.DESCRIPTION_ROW, col_index)
    label = self.parent.read_cell(SoAColumnRows.LABEL_ROW, col_index)
    epoch_name = self.parent.read_cell(SoAColumnRows.EPOCH_ROW, col_index)
    encounter_name = self.parent.read_cell(SoAColumnRows.ENCOUNTER_ROW, col_index)
    type = self.parent.read_cell(SoAColumnRows.TYPE_ROW, col_index)
    self.default_name = self.parent.read_cell(SoAColumnRows.DEFAULT_ROW, col_index)
    self.conditions = Conditons(self.parent.read_cell(SoAColumnRows.CONDITIONS_ROW, col_index))
    if encounter_name:
      encounter = self.parent.globals.cross_references.get(Encounter, encounter_name)
      if encounter:
        encounter_id = encounter.id
      else:
        self.parent._general_warning(f"Failed to find encounter with name '{encounter_name}'")
    if epoch_name:
      epoch = self.parent.globals.cross_references.get(StudyEpoch, epoch_name)
      if epoch:
        epoch_id = epoch.id
      else:
        self.parent._general_warning(f"Failed to find epoch with name '{epoch_name}'")
    try:
      if type.upper() == "ACTIVITY":
        self.item = ScheduledActivityInstance(
          id=self.parent.globals.id_manager.build_id(ScheduledActivityInstance),
          name=name,
          description=description,
          label=label,
          timelineExitId=None,
          encounterId=encounter_id,
          scheduledInstanceTimelineId=None,
          defaultConditionId=None,
          epochId=epoch_id,
          activityIds=self._add_activities()
        )
        self.parent.globals.cross_references.add(self.item.id, self.item)
        self.parent.globals.cross_references.add(self.name, self.item)
      elif type.upper() == "DECISION":
        self.item = ScheduledDecisionInstance(
          id=self.parent.globals.id_manager.build_id(ScheduledDecisionInstance),
          name=name,
          description=description,
          label=label,
          timelineExitId=None,
          scheduledInstanceTimelineId=None,
          defaultConditionId=None,
          conditionAssignments=[]
        )
        self.parent.globals.cross_references.add(self.item.id, self.item)
        self.parent.globals.cross_references.add(self.name, self.item)
      else:
        self.parent._general_warning(f"Unrecognized ScheduledInstance type: '{type}'")
    except Exception as e:
      parent._sheet_exception(f"Error raised reading sheet", e)

  def _add_activities(self):
    activities = []
    row = 0
    column = self.parent.sheet.iloc[:, self.col_index]
    for cell in column:
      if row >= SoAColumnRows.FIRST_ACTIVITY_ROW:
        activity_name = self.parent.read_cell(row, SoAColumnRows.CHILD_ACTIVITY_COL)
        if str(cell).upper() == "X":
          activity = self.parent.globals.cross_references.get(Activity, activity_name)
          if activity:
            activities.append(activity.id)
          else:
            self.parent._general_warning(f"Unable to find activity '{activity_name}' when adding to schedule instance")
      row += 1
    return activities




import logging

class Error():

  CRITICAL = logging.CRITICAL
  ERROR = logging.ERROR
  WARNING = logging.WARNING
  DEBUG = logging.DEBUG
  INFO = logging.INFO
  LABEL = { CRITICAL: "critical", ERROR: "error", WARNING: "warning", DEBUG: "debug", INFO: "info" }

  def __init__(self, message: str, level: int=ERROR):
    self.message = message
    self.level = level

  def to_log(self):
    level = self.__class__.LABEL[self.level].capitalize()  
    return f"{level}: {self.message}"

  def to_dict(self):
    return { 'message': self.message, 'level': self.__class__.LABEL[self.level].capitalize() }



from usdm_excel.base_sheet import BaseSheet
from usdm_model.abbreviation import Abbreviation
from usdm_excel.globals import Globals

class AbbreviationSheet(BaseSheet):

  SHEET_NAME = 'abbreviations'

  def __init__(self, file_path: str, globals: Globals):
    try:
      self.items = []
      super().__init__(file_path=file_path, globals=globals, sheet_name=self.SHEET_NAME, optional=True)
      if self.success:
        for index, row in self.sheet.iterrows():
          abbreviation = self.read_cell_by_name(index, 'abbreviatedText')
          text = self.read_cell_by_name(index, 'expandedText')
          params = {'abbreviatedText': abbreviation, 'expandedText': text}
          notes = self.read_cell_multiple_by_name(index, 'notes', must_be_present=False)
          item = self.create_object(Abbreviation, params)          
          if item:
            self.items.append(item)
            self.globals.cross_references.add(abbreviation, item)
            self.add_notes(item, notes) 
    except Exception as e:
      self._sheet_exception(e)



from usdm_excel.base_sheet import BaseSheet
from usdm_excel.alias import Alias
from usdm_model.study_intervention import StudyIntervention
from usdm_model.administration import Administration
from usdm_model.administration_duration import AdministrationDuration
from usdm_excel.globals import Globals

class StudyDesignInterventionSheet(BaseSheet):

  SHEET_NAME = 'studyDesignInterventions'

  def __init__(self, file_path: str, globals: Globals):
    try:
      self.items = []
      super().__init__(file_path=file_path, globals=globals, sheet_name=self.SHEET_NAME)
      self.current_name = None
      self.current_intervention = None
      for index, row in self.sheet.iterrows():
        admin_duration = self._create_administration_duration(index)
        agent_admin = self._create_administration(index, admin_duration)
        self._create_intervention(index, agent_admin)
    except Exception as e:
      self._sheet_exception(e)

  def _create_intervention(self, index, agent_admin):
    name = self.read_cell_by_name(index, 'name')
    if name and name != self.current_name:
      self.current_name = name
      params = {
        'name': name,
        'description': self.read_cell_by_name(index, 'description', must_be_present=False),
        'label': self.read_cell_by_name(index, 'label', must_be_present=False),
        'codes': self.read_other_code_cell_multiple_by_name(index, "codes"),
        'role': self.read_cdisc_klass_attribute_cell_by_name("StudyIntervention", "role", index, "role"),
        'type': self.read_cdisc_klass_attribute_cell_by_name("StudyIntervention", "type", index, "type"),
        'productDesignation': self.read_cdisc_klass_attribute_cell_by_name("StudyIntervention", "productDesignation", index, "productDesignation"),
        'minimumResponseDuration': self.read_quantity_cell_by_name(index, "minimumResponseDuration"),
        'administrations': [agent_admin]
      }
      item = self.create_object(StudyIntervention, params)
      if item:
        self.current_intervention = item
        self.items.append(item)
        self.globals.cross_references.add(name, item)
    else:
      self.current_intervention.administrations.append(agent_admin)

  def _create_administration(self, index, admin_duration):
    params = {
      'name': self.read_cell_by_name(index, 'administrationName'), 
      'description': self.read_cell_by_name(index, 'administrationDescription', must_be_present=False), 
      'label': self.read_cell_by_name(index, 'administrationLabel', must_be_present=False),
      'duration': admin_duration,
      'dose': self.read_quantity_cell_by_name(index, "administrationDose"),
      'route': Alias(self.globals).code(self.read_cdisc_klass_attribute_cell_by_name("Administration", "route", index, "administrationRoute"), []),
      'frequency': Alias(self.globals).code(self.read_cdisc_klass_attribute_cell_by_name("Administration", "frequency", index, "administrationFrequency"), [])
    }
    item = self.create_object(Administration, params)
    if item:
      self.globals.cross_references.add(item.name, item)
    return item
    
  def _create_administration_duration(self, index):
    params = {
      'description': self.read_cell_by_name(index, 'administrationDurationDescription', must_be_present=False),
      'durationWillVary': self.read_boolean_cell_by_name(index, 'administrationDurationWillVary'),
      'reasonDurationWillVary': self.read_cell_by_name(index, 'administrationDurationWillVaryReason'),
      'quantity': self.read_quantity_cell_by_name(index, 'administrationDurationQuantity')
    }
    return self.create_object(AdministrationDuration, params)



from typing import List, Literal, Union
from .api_base_model import ApiBaseModelWithIdNameLabelAndDesc
from .strength import Strength
from .code import Code

class Substance(ApiBaseModelWithIdNameLabelAndDesc):
  codes: List[Code] = []
  strengths: List[Strength] = [] # Not in API
  referenceSubstance: Union['Substance', None] = None
  instanceType: Literal['Substance']



from typing import List, Literal
from .api_base_model import ApiBaseModelWithId
from .code import Code

class CommentAnnotation(ApiBaseModelWithId):
  text: str
  codes: List[Code] = []
  instanceType: Literal['CommentAnnotation']



from usdm_excel.base_sheet import BaseSheet
from usdm_model.activity import Activity
from usdm_excel.globals import Globals
from usdm_model.comment_annotation import CommentAnnotation

class StudyDesignActivitySheet(BaseSheet):

  SHEET_NAME = 'studyDesignActivities'

  def __init__(self, file_path: str, globals: Globals):
    try:
      self.items = []
      super().__init__(file_path=file_path, globals=globals, sheet_name=self.SHEET_NAME, optional=True)
      if self.success:
        for index, row in self.sheet.iterrows():
          name = self.read_cell_by_name(index, ['activityName', 'name'])
          description = self.read_cell_by_name(index, ['activityDescription', 'description'])
          label = self.read_cell_by_name(index, 'label', default="", must_be_present=False)
          notes = self.read_cell_multiple_by_name(index, 'notes', must_be_present=False)
          item = self.create_object(Activity, {'name': name, 'description': description, 'label': label})
          if item:
            self.items.append(item)
            self.globals.cross_references.add(name, item)
            self.add_notes(item, notes)
        self.double_link(self.items, 'previousId', 'nextId')   
    except Exception as e:
      self._sheet_exception(e)




from usdm_excel.base_sheet import BaseSheet
from usdm_model.study_version import StudyVersion
from usdm_model.study_definition_document_version import StudyDefinitionDocumentVersion

class Elements():

  def __init__(self, parent: BaseSheet, study_version: StudyVersion, document_version: StudyDefinitionDocumentVersion):
    super().__init__()
    self._parent = parent
    self._study_version = study_version
    self._study_design = self._study_version.studyDesigns[0]
    self._document_version = document_version
    self._methods = [func for func in dir(self.__class__) if callable(getattr(self.__class__, func)) and not func.startswith("_")]

  def valid_method(self, name: str) -> bool:
    return name in self._methods

  def study_phase(self) -> str:
    phase = self._study_version.phase()
    results = [{'instance': phase, 'klass': 'Code', 'attribute': 'decode'}]
    return self._set_of_references(results)

  def study_short_title(self) -> str:
    title = self._study_version.get_title('Brief Study Title')
    results = [{'instance': title, 'klass': 'StudyTitle', 'attribute': 'text'}] if title else []
    return self._set_of_references(results)

  def study_full_title(self) -> str:
    title = self._study_version.get_title('Official Study Title')
    results = [{'instance': title, 'klass': 'StudyTitle', 'attribute': 'text'}] if title else []
    return self._set_of_references(results)

  def study_acronym(self) -> str:
    title = self._study_version.get_title('Study Acronym')
    results = [{'instance': title, 'klass': 'StudyTitle', 'attribute': 'text'}] if title else []
    return self._set_of_references(results)

  def study_rationale(self) -> str:
    results = [{'instance': self._study_version, 'klass': 'StudyVersion', 'attribute': 'rationale'}]
    return self._set_of_references(results)

  def study_version_identifier(self) -> str:
    results = [{'instance': self._study_version, 'klass': 'StudyVersion', 'attribute': 'versionIdentifier'}]
    return self._set_of_references(results)

  def study_identifier(self) -> str:
    identifier = self._study_version.sponsor_identifier()
    results = [{'instance': identifier, 'klass': 'StudyIdentifier', 'attribute': 'text'}]
    return self._set_of_references(results)

  def study_regulatory_identifiers(self) -> str:
    results = []
    identifiers = self._study_version.studyIdentifiers
    for identifier in identifiers:
      org = self._study_version.organization(identifier.scopeId)
      if org.type.code == 'C188863' or org.type.code == 'C93453':
        item = {'instance': identifier, 'klass': 'StudyIdentifier', 'attribute': 'text'}
        results.append(item)
    return self._set_of_references(results)

  def document_approval_date(self) -> str:
    dates = self._document_version.dateValues
    for date in dates:
      if date.type.code == 'C99903x1':
        results = [{'instance': date, 'klass': 'GovernanceDate', 'attribute': 'dateValue'}]
        return self._set_of_references(results)
    return ''
  
  def study_approval_date(self) -> str:
    dates = self._study_version.dateValues
    for date in dates:
      if date.type.code == 'C132352':
        results = [{'instance': date, 'klass': 'GovernanceDate', 'attribute': 'dateValue'}]
        return self._set_of_references(results)
    return ''

  def organization_name_and_address(self) -> str:
    identifier = self._study_version.sponsor_identifier()
    org = self._study_version.organization(identifier.scopeId)
    results = [
      {'instance': org, 'klass': 'Organization', 'attribute': 'name'},
      {'instance': org.legalAddress, 'klass': 'Address', 'attribute': 'text'},
    ]
    return self._set_of_references(results)

  def organization_address(self) -> str:
    identifier = self._study_version.sponsor_identifier()
    org = self._study_version.organization(identifier.scopeId)
    results = [
      {'instance': org.legalAddress, 'klass': 'Address', 'attribute': 'text'},
    ]
    return self._set_of_references(results)

  def organization_name(self) -> str:
    identifier = self._study_version.sponsor_identifier()
    org = self._study_version.organization(identifier.scopeId)
    results = [
      {'instance': org, 'klass': 'Organization', 'attribute': 'name'},
    ]
    return self._set_of_references(results)

  def amendment(self) -> str:
    amendments = self._study_version.amendments
    if amendments:
      results = [{'instance': amendments[-1], 'klass': 'StudyAmendment', 'attribute': 'number'}]
      return self._set_of_references(results)
    else:
      return ''

  def amendment_scopes(self) -> str:
    results = []
    amendments = self._study_version.amendments
    if amendments:
      amendment = self._study_version.amendments[-1]
      for item in amendment.geographicScopes:
        if item.type.code == "C68846":
          results = [{'instance': item.type, 'klass': 'Code', 'attribute': 'decode'}]
          return self._set_of_references(results)
        else:
          entry = {'instance': item.code.standardCode, 'klass': 'Code', 'attribute': 'decode'}
          results.append(entry)
      return self._set_of_references(results)
    else:
      return ''
    
  def no_value_for_test(self):
    return ""

  def _set_of_references(self, items):
    if items:
      return ", ".join([f'<usdm:ref klass="{item["klass"]}" id="{item["instance"].id}" attribute="{item["attribute"]}"/>' for item in items])
    else:
      return ""



from typing import Literal
from .api_base_model import ApiBaseModelWithId

class Code(ApiBaseModelWithId):
  code: str
  codeSystem: str
  codeSystemVersion: str
  decode: str
  instanceType: Literal['Code']



from typing import Union, List, Literal
from .api_base_model import ApiBaseModelWithIdNameLabelAndDesc

class AssignedPerson(ApiBaseModelWithIdNameLabelAndDesc):
  jobTitle: str
  organizationId: Union[str, None] = None
  instanceType: Literal['AssignedPerson']



import pytest
import pandas as pd
from usdm_excel.study_design_indication_sheet.study_design_indication_sheet import StudyDesignIndicationSheet
from usdm_model.code import Code


xfail = pytest.mark.xfail

def test_create(mocker, globals):
  mock_id = mocker.patch("usdm_excel.id_manager.IdManager.build_id")
  mock_id.side_effect=['Code_1', 'IndicationId_1', 'Code_2', 'IndicationId_2', 'Code_3', 'Code_4', 'IndicationId_3']
  expected_1 = Code(id='Code_1', code='X', codeSystem='SPONSOR', codeSystemVersion='', decode="Y")
  expected_2 = Code(id='Code_2', code='AAA', codeSystem='SPONSOR', codeSystemVersion='', decode="BBB")
  expected_3 = Code(id='Code_3', code='WWW', codeSystem='SPONSOR', codeSystemVersion='', decode="1234")
  expected_4 = Code(id='Code_4', code='EEE', codeSystem='SPONSOR', codeSystemVersion='', decode="3456")
  mock_code = mocker.patch("usdm_excel.cdisc_ct.CDISCCT.code_for_attribute")
  mock_code.side_effect=[expected_1, expected_2, expected_3, expected_4]
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = [['Indication 1', 'Indication One', 'Label One', 'Sponsor:X=Y'], 
          ['Indication 2', 'Indication Two', '', 'SPONSOR: AAA=BBB'], 
          ['Indication 3', 'Indication Three', '', 'SPONSOR: WWW=1234, SPONSOR: EEE=3456']]
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['name', 'description', 'label', 'codes'])
  Indications = StudyDesignIndicationSheet("", globals)
  assert len(Indications.items) == 3
  assert Indications.items[0].id == 'IndicationId_1'
  assert Indications.items[0].name == 'Indication 1'
  assert Indications.items[1].id == 'IndicationId_2'
  assert Indications.items[1].description == 'Indication Two'
  assert Indications.items[1].codes == [expected_2]
  assert Indications.items[2].id == 'IndicationId_3'
  assert Indications.items[2].codes == [expected_3, expected_4]
  
def test_create_empty(mocker, globals):
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = []
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['studyIndicationName', 'studyIndicationDescription', 'studyIndicationType'])
  Indications = StudyDesignIndicationSheet("", globals)
  assert len(Indications.items) == 0

def test_read_cell_by_name_error(mocker, globals):
  mock_error = mocker.patch("usdm_excel.errors_and_logging.errors.Errors.add")
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = [['Indication 1', 'Indication One']]
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['name', 'description'])
  Indications = StudyDesignIndicationSheet("", globals)
  mock_error.assert_called()
  assert mock_error.call_args[0][0] == "studyDesignIndications"
  assert mock_error.call_args[0][1] == None
  assert mock_error.call_args[0][2] == None
  assert mock_error.call_args[0][3] == "Exception. Error [Failed to detect column(s) 'codes' in sheet] while reading sheet 'studyDesignIndications'. See log for additional details."
  



from usdm_excel.base_sheet import BaseSheet
from usdm_excel.study_soa_v2_sheet.soa_activities import SoAActivities
from usdm_excel.study_soa_v2_sheet.scheduled_instances import ScheduledInstances
from usdm_model.scheduled_instance import ScheduledActivityInstance, ScheduledDecisionInstance
from usdm_model.schedule_timeline import ScheduleTimeline
from usdm_excel.globals import Globals

class StudySoAV2Sheet(BaseSheet):

  NAME_ROW = 0
  DESCRIPTION_ROW = 1
  CONDITION_ROW = 2
  PARAMS_DATA_COL = 1

  def __init__(self, file_path: str, globals: Globals, sheet_name: str, main: bool=False, require: dict={}):
    try:
      self.name = ""
      self.description = ""
      self.condition = ""
      self.timeline = None
      self.main_timeline = main
      self.activities = []
      self.biomedical_concepts = []
      self.biomedical_concept_surrogates = []
      super().__init__(file_path=file_path, globals=globals, sheet_name=sheet_name, header=None, require=require)
      self._process_sheet()
      self._raw_activities = SoAActivities(self) # Order important, activities then instances
      self._raw_instances = ScheduledInstances(self)
      self.activities, self.biomedical_concepts, self.biomedical_concept_surrogates = self._raw_activities.group_and_link()
      self._raw_activities.set_parents()
      self.timeline = self._add_timeline(self.name, self.description, self.condition, self._raw_instances.instances, self._raw_instances.exits)

    except Exception as e:
      self._sheet_exception(e)

  def check_timing_references(self, timings, timing_check):
    timing_set = []
    for instance in self._raw_instances.items:
      item = instance.item
      #print(f"TIMING1: {item.id}, {instance.name}")
      if isinstance(item, ScheduledActivityInstance):
        found = False
        for timing in timings:
          ids = [timing.relativeFromScheduledInstanceId, timing.relativeToScheduledInstanceId]
          #print(f"TIMING2: {timing.name}, {ids}")
          if item.id in ids:
            #print(f"TIMING3: found")
            found = True
            if not timing_check[timing.name]:
              timing_check[timing.name] = self.name
              timing_set.append(timing)
            elif timing_check[timing.name] == self.name:
              pass
            else:
              self._general_warning(f"Duplicate use of timing with name '{timing.name}' across timelines detected")
            #break
        if not found:
          self._general_warning(f"Unable to find timing reference for instance with name '{instance.name}'")
    return timing_set
  
  def timing_match(self, ref):
    return self._raw_instances.match(ref)

  def _process_sheet(self):
    for rindex in range(self.NAME_ROW, self.CONDITION_ROW + 1):
      if rindex == self.NAME_ROW:
        self.name = self.read_cell(rindex, self.PARAMS_DATA_COL)
      elif rindex == self.DESCRIPTION_ROW:
        self.description = self.read_cell(rindex, self.PARAMS_DATA_COL)
      elif rindex == self.CONDITION_ROW:
        self.condition = self.read_cell(rindex, self.PARAMS_DATA_COL)
      else:
        pass

  def _add_timeline(self, name, description, condition, instances, exit):
    try:
      timeline = ScheduleTimeline(
        id=self.globals.id_manager.build_id(ScheduleTimeline),
        mainTimeline=self.main_timeline,
        name=name,
        description=description,
        label=name,
        entryCondition=condition,
        entryId=instances[0].id,
        exits=exit,
        instances=instances
      )
      self.globals.cross_references.add(timeline.name, timeline)
      return timeline
    except Exception as e:
      self._general_exception(f"Failed to create ScheduleTimeline object", e)
      return None







from typing import List, Literal, Union
from .api_base_model import ApiBaseModelWithId
from .identifier import *
from .study_definition_document_version import *
from .alias_code import *
from .code import Code as genericCode
from .study_design import *
from .governance_date import GovernanceDate
from .study_amendment import StudyAmendment
from .study_title import StudyTitle
from .eligibility_criterion import EligibilityCriterion
from .narrative_content import NarrativeContentItem
from .comment_annotation import CommentAnnotation
from .abbreviation import Abbreviation
from .study_role import StudyRole
from .organization import Organization
from .administrable_product import AdministrableProduct

class StudyVersion(ApiBaseModelWithId):
  versionIdentifier: str
  rationale: str
  studyType: Union[genericCode, None] = None
  studyPhase: Union[AliasCode, None] = None
  documentVersionIds: List[str] = []
  dateValues: List[GovernanceDate] = []
  amendments: List[StudyAmendment] = []
  businessTherapeuticAreas: List[Code] = []
  studyIdentifiers: List[StudyIdentifier]
  referenceIdentifiers: List[ReferenceIdentifier] = []
  studyDesigns: List[StudyDesign] = []
  titles: List[StudyTitle]
  criteria: List[EligibilityCriterion] = [] # Not in API
  narrativeContentItems: List[NarrativeContentItem] = []
  abbreviations: List[Abbreviation] = []
  roles: List[StudyRole] = []
  organizations: List[Organization] = []
  administrableProducts: List[AdministrableProduct] = []
  notes: List[CommentAnnotation] = []
  instanceType: Literal['StudyVersion']

  def get_title(self, title_type):
    for title in self.titles:
      if title.type.decode == title_type:
        return title
    return None

  def sponsor_identifier(self):
    for identifier in self.studyIdentifiers:
      org = self.organization(identifier.scopeId)
      if org and org.type.code == 'C70793':
        return identifier
    return None

  def phase(self):
    return self.studyPhase.standardCode
  
  def organization(self, id: str) -> Organization:
    return next((x for x in self.organizations if x.id == id), None)


from typing import List, Literal
from .alias_code import AliasCode
from .api_base_model import ApiBaseModelWithIdNameAndLabel
from .biomedical_concept_property import BiomedicalConceptProperty
from .comment_annotation import CommentAnnotation

class BiomedicalConcept(ApiBaseModelWithIdNameAndLabel):
  synonyms: List[str] = []
  reference: str
  properties: List[BiomedicalConceptProperty] = []
  code: AliasCode
  notes: List[CommentAnnotation] = []
  instanceType: Literal['BiomedicalConcept']



from typing import List, Literal, Union
from .api_base_model import ApiBaseModelWithIdNameLabelAndDesc
from .code import Code
from .comment_annotation import CommentAnnotation

class StudyEpoch(ApiBaseModelWithIdNameLabelAndDesc):
  type: Code
  previousId: Union[str, None] = None
  nextId: Union[str, None] = None
  notes: List[CommentAnnotation] = []
  instanceType: Literal['StudyEpoch']



import pytest
import pandas as pd
from usdm_excel.study_design_objective_endpoint_sheet.study_design_objective_endpoint_sheet import StudyDesignObjectiveEndpointSheet
from usdm_model.api_base_model import ApiBaseModelWithId


xfail = pytest.mark.xfail

def test_create(mocker, globals):
  #mock_cross_ref = mocker.patch("usdm_excel.cross_ref.CrossRef.get")
  #mock_cross_ref.side_effect=[ApiBaseModelWithId(id="1"), ApiBaseModelWithId(id="2"), ApiBaseModelWithId(id="3"), ApiBaseModelWithId(id="4"), ApiBaseModelWithId(id="5")]
  mock_present = mocker.patch("usdm_excel.base_sheet.BaseSheet._sheet_present")
  mock_present.side_effect=[True]
  mock_id = mocker.patch("usdm_excel.id_manager.IdManager.build_id")
  mock_id.side_effect=[
    'Code_1', 'Code_2', 'ObjId_1', 'EndId_1', 'Code_3', 'EndId_2', 
    'Code_4', 'Code_5', 'ObjId_2', 'EndId_3', 
    'Code_6', 'Code_7', 'ObjId_3', 'EndId_4', 'Code_8', 'EndId_5'
  ]
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = [
    ['Obj 1','Obj Desc 1', 'Obj Label 1', 'Obj Text 1', 'Primary Objective',   'End 1', 'End Desc 1', 'End Label 1', 'End Text 1', '', 'Primary Endpoint'], 
    ['','', '', '', '',                                                        'End 2', 'End Desc 2', 'End Label 2', 'End Text 2', '', 'Primary Endpoint'], 
    ['Obj 2','Obj Desc 2', 'Obj Label 2', 'Obj Text 2', 'Secondary Objective', 'End 3', 'End Desc 3', 'End Label 3', 'End Text 3', '', 'Secondary Endpoint'], 
    ['Obj 3','Obj Desc 3', 'Obj Label 3', 'Obj Text 3', 'Secondary Objective', 'End 4', 'End Desc 4', 'End Label 4', 'End Text 4', '', 'Secondary Endpoint'], 
    ['','', '', '', '',                                                        'End 5', 'End Desc 5', 'End Label 5', 'End Text 5', '', 'Secondary Endpoint']
  ]
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=[
    'objectiveName', 'objectiveDescription', 'objectiveLabel', 'objectiveText', 
    'objectiveLevel', 'endpointName', 'endpointDescription', 'endpointLabel', 'endpointText', 'endpointPurpose', 'endpointLevel']
  )
  items = StudyDesignObjectiveEndpointSheet("", globals)
  assert len(items.objectives) == 3
  assert items.objectives[0].id == 'ObjId_1'
  assert items.objectives[0].name == 'Obj 1'
  assert items.objectives[0].description == 'Obj Desc 1'
  assert items.objectives[0].label == 'Obj Label 1'
  assert items.objectives[0].endpoints[0].name == 'End 1'
  assert items.objectives[0].endpoints[1].name == 'End 2'
  assert items.objectives[1].id == 'ObjId_2'
  assert items.objectives[1].name == 'Obj 2'
  assert items.objectives[1].endpoints[0].name == 'End 3'
  assert items.objectives[2].id == 'ObjId_3'
  assert items.objectives[2].name == 'Obj 3'
  assert items.objectives[2].endpoints[0].name == 'End 4'
  assert items.objectives[2].endpoints[1].name == 'End 5'
  
def test_create_empty(mocker, globals):
  mock_present = mocker.patch("usdm_excel.base_sheet.BaseSheet._sheet_present")
  mock_present.side_effect=[True]
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = []
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=[
    'objectiveName', 'objectiveDescription', 'objectiveLabel', 'objectiveText', 
    'objectiveLevel', 'endpointName', 'endpointDescription', 'endpointLabel', 'endpointText', 'endpointPurpose', 'endpointLevel']
  )
  items = StudyDesignObjectiveEndpointSheet("", globals)
  assert len(items.objectives) == 0

def test_read_cell_by_name_error(mocker, globals):
  
  call_parameters = []
  
  def my_add(sheet, row, column, message, level=10):
    call_parameters.append((sheet, row, column, message, level))
    return None

  mock_present = mocker.patch("usdm_excel.base_sheet.BaseSheet._sheet_present")
  mock_present.side_effect=[True]
  mock_error = mocker.patch("usdm_excel.errors_and_logging.errors.Errors.add", side_effect=my_add)
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = [
    ['Obj 1','Obj Desc 1', 'Obj Label 1', 'Obj Text 1', 'Primary Objective', 
     'End 1', 'End Desc 1', 'End Label 1', 'End Text Purpose 1', 'Primary Endpoint']
 ]
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=[
    'objectiveName', 'objectiveDescription', 'objectiveLabel', 'objectiveText', 'objectiveLevel', 
    'endpointName', 'endpointDescription', 'endpointLabel', 'endpointPurpose', 'endpointLevel']
  )
  items = StudyDesignObjectiveEndpointSheet("", globals)
  mock_error.assert_called()
  assert call_parameters[0] == ("studyDesignOE", 1, -1, "Error attempting to read cell 'endpointText'. Exception: Failed to detect column(s) 'endpointText' in sheet", 40)
  



import pytest
import pandas as pd

xfail = pytest.mark.xfail

from usdm_excel.study_sheet.study_sheet import StudySheet
from usdm_excel.base_sheet import BaseSheet

@xfail
def test_create(mocker, globals):
  assert 0
  # mocked_open = mocker.mock_open(read_data="File")
  # mocker.patch("builtins.open", mocked_open)
  # mock_read = mocker.patch("pandas.read_excel")
  # data = { 
  #   'col_1': 
  #     [ 'studyTitle', 'studyVersion', 'studyType', 'studyPhase', 'studyAcronym', 'studyRationale', 
  #       'businessTherapueticAreas', '', 'briefTitle', 'Something brief' ],
  #   'col_2': 
  #     [ 'title', '1', 'Interventional Study', 'C15602', 'BEAST', 'Damn good', 
  #       '', '', 'officalTitle', 'Very offical' ],
  #   'col_3': [ '', '', '', '', '', '', '', '', 'publicTitle', 'Open' ],
  #   'col_4': [ '', '', '', '', '', '', '', '', 'scientificTitle', 'Blah blah blah' ],
  #   'col_5': [ '', '', '', '', '', '', '', '', 'protocolVersion', '1' ],
  #   'col_6': [ '', '', '', '', '', '', '', '', 'protocolAmendment', 'A' ],
  #   'col_7': [ '', '', '', '', '', '', '', '', 'protocolEffectiveDate', '2010-01-01 00:00:00' ],
  #   'col_8': [ '', '', '', '', '', '', '', '', 'protocolStatus', 'draft' ],
  #   }
  # mock_read.return_value = pd.DataFrame.from_dict(data)
  # study = StudySheet("")
  # assert(study.study.studyTitle) == 'title'
  # assert(study.study.studyVersion) == '1'

